/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
MHQuake 6
New options menu drawing system - more easily extensible

to add a new item to a menu, do the following:

1) declare a cvar and do whatever you want to with it elsewhere in the game in the usual fashion

2) declare it as an extern in this file

3) select a menu to put it in

4) use a m_menuitem_draw function *anywhere* in the menu - mhquake will automatically put it in
   the right place for you and look after the slider adjustment, etc.  no worries.  cool.

here's how the parameters of m_menuitem_draw work:

void M_MenuItem_Draw (float menu_cvar, mitem_type_t mitem_type, char *mitem_string,
					  char *m_switch_1, char *m_switch_2, char *m_switch_3, cvar_t *menu_store,
					  float scale_min, float scale_max, menu_store_t *store, reload_t reload,
					  float step, (void (*submenu_func) ())

menu_cvar:		the current value of the relevant cvar adjusted to a 0 to 1 scale - use 0.0 for
				types that don't require use of a cvar.

mitem_type:		an enum - one of the following:
				on_off:			a standard 2 option checkbox
				on_off2:		as above except with custom values rather than 0 and 1
				three_item:		like above only with 3 options
				slider:			a standard slider
				scale:			like above only prints the value rather than uses a slider
				sub_menu:		goes to a sub menu - in reality, this can be used to call any
								function you want from an options menu - very powerful, and
								very very scary...
				spacer:			use as a spacer in the menu list - you must provide your own
								string (it can be " " or anything you want)

				all of these need a cvar except the last 2, where you should use temp1.

mitem_string:	what you want printed in the menu text (use " " for spacers)

m_switch_1:		option 1 for standard and three-item checkboxes (use "MH" for other types) - can
				also be used for spacer types where you want something to display at the right.

m_switch_2:		option 2 for standard and three-item checkboxes (use "MH" for other types)

m_switch_3:		option 3 for three-item checkboxes (use "MH" if a standard checkbox or other type)

menu_store:		a pointer to the cvar (use &cvar_name) - required as well as the value because
				we must pass an adjusted value to the function - use &temp1 for types that don't
				require use of a cvar.

scale_min:		the minimum value for slider and scale types - use 0.0 for other types - for
				on_off2 types, this is the value of the 'off' state.

scale_max:		the maximum value for slider and scale types - use 0.0 for other types - for
				on_off2 types, this is the value of the 'on' state.

store:			which array the menu item is in - eg, &general_store[0] for the general menu.

				&options_store[0]:	goes on the main options menu
				&general_store[0]:	goes on the general view menu
				&basic_store[0]:	goes on the basic speed vs. ugliness modification options menu
				&static_store[0]:	goes on the static lighting menu
				&dynamic_store[0]:	goes on the dynamic lighting menu
				&candy_store[0]:	goes on the basic eyecandy menu
				&hardcore_store[0]:	goes on the hardcore cpu-killing eyecandy options menu
				&fog_store[0]:		goes on the global fog menu

reload:			informs if the map needs to be reloaded after changing the option (set to
				reload_map) or if quake needs to be restarted (set to restart_quake) - normally
				will be no_reload.

step:			the amount slider and scale types are incremented/decremented: use 0.0 for
				other types.

submenu_func:	the function called by a submenu (very powerful this, use with care!!!) - use
				M_FuncNull if there is none.  if you don't understand function pointers, don't
				touch this - although it should be easy to see how they work from having a look.

since most of these are pointers (and the ones that aren't are small enough), the stack shouldn't
take too much of a hammering (nothing's crashed on me yet anyway...) - in fact, it uses a mighty
52 bytes of memory... so there.

extensibility example - i added the fps counter to the general view menu by adding the following
lines of code:

(among the other extern cvar declarations:)
extern cvar_t cl_fpscount;

(in the general view menu draw function:)
M_MenuItem_Draw (cl_fpscount.value, on_off, "FPS Counter", "On", "Off", "MH", &cl_fpscount,
	0.0, 0.0, &general_store[0], no_reload, 0.0, M_FuncNull);

Totally painless.

If you want to move an item to a different place in the same menu, it's just a simple matter of
cutting an pasting the relevant line.  If you want to move an item to a different place in a
different menu - (believe it or not...) cut and paste as before - too easy.
*/

#include "quakedef.h"
#include "winquake.h"

enum
{
    m_none,
    m_main,
    m_singleplayer,
    m_load,
    m_save,
    m_multiplayer,
    m_setup,
    m_net,
    m_options,
    m_keys,
    m_help,
    m_quit,
    m_serialconfig,
    m_modemconfig,
    m_lanconfig,
    m_gameoptions,
    m_search,
    m_slist,
    m_options_general,
    m_options_basic,
    m_options_static,
    m_options_dynamic,
    m_options_candy,
    m_options_hardcore,
    m_options_fog
} m_state;


// structures and stuff for the new submenu options system
typedef enum
{
    on_off,
    on_off2,
    three_item,
    slider,
    scale,
    sub_menu,
    spacer
} mitem_type_t;


typedef enum
{
    no_reload,
    reload_map,
    restart_quake
} reload_t;


typedef struct menu_store_s
{
    cvar_t *cvar_store;
    float min_scale;
    float max_scale;
    mitem_type_t item_type;
    reload_t requires_reload;
    float slider_step;
    void (*submenu_func) ();
} menu_store_t;

void (*Previous_Options_Menu) ();

// 25 gives us plenty of room for expansion - static arrays are easier to manage in this instance
// than dynamic ones, so we'll stick with them.  you may thing i'm lazy, but since this requires
// an extra 224 bytes, i reckon your RAM can handle it.  if you're feeling brave, try increasing
// it to 50 for a whopping 448 bytes!!!
#define MAX_MENU_STORE 25

menu_store_t options_store[MAX_MENU_STORE];
menu_store_t general_store[MAX_MENU_STORE];
menu_store_t basic_store[MAX_MENU_STORE];
menu_store_t static_store[MAX_MENU_STORE];
menu_store_t dynamic_store[MAX_MENU_STORE];
menu_store_t candy_store[MAX_MENU_STORE];
menu_store_t hardcore_store[MAX_MENU_STORE];
menu_store_t fog_store[MAX_MENU_STORE];


void M_MenuItem_Draw (float menu_cvar, mitem_type_t mitem_type, char *mitem_string,
                      char *m_switch_1, char *m_switch_2, char *m_switch_3, cvar_t *menu_store,
                      float scale_min, float scale_max, menu_store_t *store, reload_t reload,
                      float step, void (*submenu_func) ());

void M_Menu_Main_f (void);
void M_Menu_SinglePlayer_f (void);
void M_Menu_Load_f (void);
void M_Menu_Save_f (void);
void M_Menu_MultiPlayer_f (void);
void M_Menu_Setup_f (void);
void M_Menu_Net_f (void);
void M_Menu_Options_f (void);
void M_Menu_Keys_f (void);
void M_Menu_General_f (void);
void M_Menu_Basic_f (void);
void M_Menu_Static_f (void);
void M_Menu_Dynamic_f (void);
void M_Menu_Candy_f (void);
void M_Menu_HardCore_f (void);
void M_Menu_Fog_f (void);
void M_Menu_Help_f (void);
void M_Menu_Quit_f (void);
void M_Menu_SerialConfig_f (void);
void M_Menu_ModemConfig_f (void);
void M_Menu_LanConfig_f (void);
void M_Menu_GameOptions_f (void);
void M_Menu_Search_f (void);
void M_Menu_ServerList_f (void);

void M_Main_Draw (void);
void M_SinglePlayer_Draw (void);
void M_Load_Draw (void);
void M_Save_Draw (void);
void M_MultiPlayer_Draw (void);
void M_Setup_Draw (void);
void M_Net_Draw (void);
void M_Options_Draw (int draw_cursor, menu_store_t *current_store);
void M_Keys_Draw (void);
void M_General_Draw (menu_store_t *current_store);
void M_Basic_Draw (menu_store_t *current_store);
void M_Static_Draw (menu_store_t *current_store);
void M_Dynamic_Draw (menu_store_t *current_store);
void M_Candy_Draw (menu_store_t *current_store);
void M_HardCore_Draw (menu_store_t *current_store);
void M_Fog_Draw (menu_store_t *current_store);
void M_Help_Draw (void);
void M_Quit_Draw (void);
void M_SerialConfig_Draw (void);
void M_ModemConfig_Draw (void);
void M_LanConfig_Draw (void);
void M_GameOptions_Draw (void);
void M_Search_Draw (void);
void M_ServerList_Draw (void);

void M_Main_Key (int key);
void M_SinglePlayer_Key (int key);
void M_Load_Key (int key);
void M_Save_Key (int key);
void M_MultiPlayer_Key (int key);
void M_Setup_Key (int key);
void M_Net_Key (int key);
void M_Keys_Key (int key);
int M_SubMenu_Key (int key, int submenu_options_cursor, menu_store_t *store);
void M_Help_Key (int key);
void M_Quit_Key (int key);
void M_SerialConfig_Key (int key);
void M_ModemConfig_Key (int key);
void M_LanConfig_Key (int key);
void M_GameOptions_Key (int key);
void M_Search_Key (int key);
void M_ServerList_Key (int key);

qboolean	m_entersound;		// play after drawing a frame, so caching
// won't disrupt the sound
qboolean	m_recursiveDraw;
int			m_return_state;
qboolean	m_return_onerror;
char		m_return_reason [32];

#define StartingGame	(m_multiplayer_cursor == 1)
#define JoiningGame		(m_multiplayer_cursor == 0)
#define SerialConfig	(m_net_cursor == 0)
#define DirectConfig	(m_net_cursor == 1)
#define	IPXConfig		(m_net_cursor == 2)
#define	TCPIPConfig		(m_net_cursor == 3)

void M_ConfigureNetSubsystem(void);

// do nothing - for null function pointers
void M_FuncNull (void)
{
    return;
}

/*
================
M_DrawCharacter

Draws one solid graphics character
================
*/
void M_DrawCharacter (int cx, int line, int num)
{
    Draw_Character (cx + 8 + ((vid.width - 320)>>1), line, num);
}

void M_Print (int cx, int cy, char *str)
{
    while (*str)
    {
        M_DrawCharacter (cx + 8, cy, (*str)+128);
        str++;
        cx += 8;
    }
}

void M_PrintWhite (int cx, int cy, char *str)
{
    while (*str)
    {
        M_DrawCharacter (cx + 8, cy, *str);
        str++;
        cx += 8;
    }
}

void M_DrawTransPic (int x, int y, qpic_t *pic)
{
    Draw_TransPic (x + 8 + ((vid.width - 320)>>1), y, pic);
}

void M_DrawTransPic2 (int x, int y, qpic_t *pic)
{
    Draw_TransPic (x + ((vid.width - 320)>>1), y, pic);
}

void M_DrawPic (int x, int y, qpic_t *pic)
{
    Draw_Pic (x + 8 + ((vid.width - 320)>>1), y, pic);
}

byte identityTable[256];
byte translationTable[256];

void M_BuildTranslationTable(int top, int bottom)
{
    int		j;
    byte	*dest, *source;

    for (j = 0; j < 256; j++)
        identityTable[j] = j;
    dest = translationTable;
    source = identityTable;
    memcpy (dest, source, 256);

    if (top < 128)	// the artists made some backwards ranges.  sigh.
        memcpy (dest + TOP_RANGE, source + top, 16);
    else
        for (j=0 ; j<16 ; j++)
            dest[TOP_RANGE+j] = source[top+15-j];

    if (bottom < 128)
        memcpy (dest + BOTTOM_RANGE, source + bottom, 16);
    else
        for (j=0 ; j<16 ; j++)
            dest[BOTTOM_RANGE+j] = source[bottom+15-j];
}


void M_DrawTransPicTranslate (int x, int y, qpic_t *pic)
{
    Draw_TransPicTranslate (x + ((vid.width - 320)>>1), y, pic, translationTable);
}


void Draw_Image (int image_texnum, float x0, float y0, float x1, float y1);
void Draw_TextBox (int x, int y, int width, int lines);

void M_DrawTextBox (int x, int y, int width, int lines)
{
    Draw_TextBox (x + ((vid.width - 320)>>1), y, width, lines);
}

//=============================================================================

int m_save_demonum;

/*
================
M_ToggleMenu_f
================
*/
void M_ToggleMenu_f (void)
{
    m_entersound = true;

    if (key_dest == key_menu)
    {
        if (m_state != m_main)
        {
            M_Menu_Main_f ();
            return;
        }
        key_dest = key_game;
        m_state = m_none;
        return;
    }
    if (key_dest == key_console)
    {
        Con_ToggleConsole_f ();
    }
    else
    {
        M_Menu_Main_f ();
    }
}


//=============================================================================
/* MAIN MENU */

int	m_main_cursor;
#define	MAIN_ITEMS	5

void M_Menu_Main_f (void)
{
    if (key_dest != key_menu)
    {
        m_save_demonum = cls.demonum;
        cls.demonum = -1;
    }

    key_dest = key_menu;
    m_state = m_main;
    m_entersound = true;
}


void M_Main_Draw (void)
{
    int		f;
    qpic_t	*p;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/ttl_main.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);
    M_DrawTransPic (72, 32, Draw_CachePic ("gfx/mainmenu.lmp") );

    f = (int)(host_time * 10)%6;

    M_DrawTransPic (46, 32 + m_main_cursor * 20,Draw_CachePic( va("gfx/menudot%i.lmp", f+1 ) ) );
}


void M_Main_Key (int key)
{
    switch (key)
    {
    case K_ESCAPE:
        key_dest = key_game;
        m_state = m_none;
        cls.demonum = m_save_demonum;
        if (cls.demonum != -1 && !cls.demoplayback && cls.state != ca_connected)
            CL_NextDemo ();
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        if (++m_main_cursor >= MAIN_ITEMS)
            m_main_cursor = 0;
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        if (--m_main_cursor < 0)
            m_main_cursor = MAIN_ITEMS - 1;
        break;

    case K_ENTER:
        m_entersound = true;

        switch (m_main_cursor)
        {
        case 0:
            M_Menu_SinglePlayer_f ();
            break;

        case 1:
            M_Menu_MultiPlayer_f ();
            break;

        case 2:
            M_Menu_Options_f ();
            break;

        case 3:
            M_Menu_Help_f ();
            break;

        case 4:
            M_Menu_Quit_f ();
            break;
        }
    }
}

//=============================================================================
/* SINGLE PLAYER MENU */

int	m_singleplayer_cursor;
#define	SINGLEPLAYER_ITEMS	3


void M_Menu_SinglePlayer_f (void)
{
    key_dest = key_menu;
    m_state = m_singleplayer;
    m_entersound = true;
}


void M_SinglePlayer_Draw (void)
{
    int		f;
    qpic_t	*p;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/ttl_sgl.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);
    M_DrawTransPic (72, 32, Draw_CachePic ("gfx/sp_menu.lmp") );

    f = (int)(host_time * 10)%6;

    M_DrawTransPic (46, 32 + m_singleplayer_cursor * 20,Draw_CachePic( va("gfx/menudot%i.lmp", f+1 ) ) );
}


void M_SinglePlayer_Key (int key)
{
    switch (key)
    {
    case K_ESCAPE:
        M_Menu_Main_f ();
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        if (++m_singleplayer_cursor >= SINGLEPLAYER_ITEMS)
            m_singleplayer_cursor = 0;
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        if (--m_singleplayer_cursor < 0)
            m_singleplayer_cursor = SINGLEPLAYER_ITEMS - 1;
        break;

    case K_ENTER:
        m_entersound = true;

        switch (m_singleplayer_cursor)
        {
        case 0:
            if (sv.active)
                if (!SCR_ModalMessage("Are you sure you want to\nstart a new game?\n"))
                    break;

            key_dest = key_game;

            if (sv.active)
                Cbuf_AddText ("disconnect\n");

            Cbuf_AddText ("maxplayers 1\n");
            Cbuf_AddText ("map start\n");
            break;

        case 1:
            M_Menu_Load_f ();
            break;

        case 2:
            M_Menu_Save_f ();
            break;
        }
    }
}

//=============================================================================
/* LOAD/SAVE MENU */

int		load_cursor;		// 0 < load_cursor < MAX_SAVEGAMES

#define	MAX_SAVEGAMES		12
char	m_filenames[MAX_SAVEGAMES][SAVEGAME_COMMENT_LENGTH+1];
int		loadable[MAX_SAVEGAMES];

void M_ScanSaves (void)
{
    int		i, j;
    char	name[MAX_OSPATH];
    FILE	*f;
    int		version;

    sprintf (name, "%s/save", com_gamedir);
    Sys_mkdir (name);

    for (i=0 ; i<MAX_SAVEGAMES ; i++)
    {
        strcpy (m_filenames[i], "--- UNUSED SLOT ---");
        loadable[i] = false;
        sprintf (name, "%s/save/s%i.sav", com_gamedir, i);

        f = fopen (name, "r");

        if (!f)
            continue;

        fscanf (f, "%i\n", &version);
        fscanf (f, "%79s\n", name);

        strncpy (m_filenames[i], name, sizeof(m_filenames[i])-1);

        // change _ back to space
        for (j=0 ; j<SAVEGAME_COMMENT_LENGTH ; j++)
            if (m_filenames[i][j] == '_')
                m_filenames[i][j] = ' ';

        loadable[i] = true;
        fclose (f);
    }
}


void M_Menu_Load_f (void)
{
    m_entersound = true;
    m_state = m_load;
    key_dest = key_menu;
    M_ScanSaves ();
}


void M_Menu_Save_f (void)
{
    if (!sv.active)
        return;
    if (cl.intermission)
        return;
    if (svs.maxclients != 1)
        return;
    m_entersound = true;
    m_state = m_save;
    key_dest = key_menu;
    M_ScanSaves ();
}


void M_Load_Draw (void)
{
    int		i;
    qpic_t	*p;

    p = Draw_CachePic ("gfx/p_load.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    for (i=0 ; i< MAX_SAVEGAMES; i++)
        M_Print (16, 32 + 8*i, m_filenames[i]);

    // line cursor
    M_DrawCharacter (8, 32 + load_cursor*8, 12+((int)(realtime*4)&1));
}


void M_Save_Draw (void)
{
    int		i;
    qpic_t	*p;

    p = Draw_CachePic ("gfx/p_save.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    for (i=0 ; i<MAX_SAVEGAMES ; i++)
        M_Print (16, 32 + 8*i, m_filenames[i]);

    // line cursor
    M_DrawCharacter (8, 32 + load_cursor*8, 12+((int)(realtime*4)&1));
}


void M_Load_Key (int k)
{
    switch (k)
    {
    case K_ESCAPE:
        M_Menu_SinglePlayer_f ();
        break;

    case K_ENTER:
        S_LocalSound ("misc/menu2.wav");
        if (!loadable[load_cursor])
            return;
        m_state = m_none;
        key_dest = key_game;

        // Host_Loadgame_f can't bring up the loading plaque because too much
        // stack space has been used, so do it now
        SCR_BeginLoadingPlaque ();

        // issue the load command
        Cbuf_AddText (va ("load s%i\n", load_cursor) );
        return;

    case K_UPARROW:
    case K_LEFTARROW:
        S_LocalSound ("misc/menu1.wav");
        load_cursor--;
        if (load_cursor < 0)
            load_cursor = MAX_SAVEGAMES-1;
        break;

    case K_DOWNARROW:
    case K_RIGHTARROW:
        S_LocalSound ("misc/menu1.wav");
        load_cursor++;
        if (load_cursor >= MAX_SAVEGAMES)
            load_cursor = 0;
        break;
    }
}


void M_Save_Key (int k)
{
    switch (k)
    {
    case K_ESCAPE:
        M_Menu_SinglePlayer_f ();
        break;

    case K_ENTER:
        m_state = m_none;
        key_dest = key_game;
        Cbuf_AddText (va("save s%i\n", load_cursor));
        return;

    case K_UPARROW:
    case K_LEFTARROW:
        S_LocalSound ("misc/menu1.wav");
        load_cursor--;
        if (load_cursor < 0)
            load_cursor = MAX_SAVEGAMES-1;
        break;

    case K_DOWNARROW:
    case K_RIGHTARROW:
        S_LocalSound ("misc/menu1.wav");
        load_cursor++;
        if (load_cursor >= MAX_SAVEGAMES)
            load_cursor = 0;
        break;
    }
}

//=============================================================================
/* MULTIPLAYER MENU */

int	m_multiplayer_cursor;
#define	MULTIPLAYER_ITEMS	3


void M_Menu_MultiPlayer_f (void)
{
    key_dest = key_menu;
    m_state = m_multiplayer;
    m_entersound = true;
}


void M_MultiPlayer_Draw (void)
{
    int		f;
    qpic_t	*p;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);
    M_DrawTransPic (72, 32, Draw_CachePic ("gfx/mp_menu.lmp") );

    f = (int)(host_time * 10)%6;

    M_DrawTransPic (46, 32 + m_multiplayer_cursor * 20,Draw_CachePic( va("gfx/menudot%i.lmp", f+1 ) ) );

    if (serialAvailable || ipxAvailable || tcpipAvailable)
        return;

    M_PrintWhite ((320/2) - ((27*8)/2), 148, "No Communications Available");
}


void M_MultiPlayer_Key (int key)
{
    switch (key)
    {
    case K_ESCAPE:
        M_Menu_Main_f ();
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        if (++m_multiplayer_cursor >= MULTIPLAYER_ITEMS)
            m_multiplayer_cursor = 0;
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        if (--m_multiplayer_cursor < 0)
            m_multiplayer_cursor = MULTIPLAYER_ITEMS - 1;
        break;

    case K_ENTER:
        m_entersound = true;
        switch (m_multiplayer_cursor)
        {
        case 0:
            if (serialAvailable || ipxAvailable || tcpipAvailable)
                M_Menu_Net_f ();
            break;

        case 1:
            if (serialAvailable || ipxAvailable || tcpipAvailable)
                M_Menu_Net_f ();
            break;

        case 2:
            M_Menu_Setup_f ();
            break;
        }
    }
}

//=============================================================================
/* SETUP MENU */

int		setup_cursor = 4;
int		setup_cursor_table[] = {40, 56, 80, 104, 140};

char	setup_hostname[16];
char	setup_myname[16];
int		setup_oldtop;
int		setup_oldbottom;
int		setup_top;
int		setup_bottom;

#define	NUM_SETUP_CMDS	5

void M_Menu_Setup_f (void)
{
    key_dest = key_menu;
    m_state = m_setup;
    m_entersound = true;
    strcpy(setup_myname, cl_name.string);
    strcpy(setup_hostname, hostname.string);
    setup_top = setup_oldtop = ((int)cl_color.value) >> 4;
    setup_bottom = setup_oldbottom = ((int)cl_color.value) & 15;
}


void M_Setup_Draw (void)
{
    qpic_t	*p;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    M_Print (48, 40, "Hostname");
    M_DrawTextBox (168, 32, 16, 1);
    M_Print (160, 40, setup_hostname);

    M_Print (48, 56, "Your name");
    M_DrawTextBox (168, 48, 16, 1);
    M_Print (160, 56, setup_myname);

    M_Print (48, 80, "Shirt color");
    M_Print (48, 104, "Pants color");

    M_DrawTextBox (56, 132, 14, 1);
    M_Print (48, 140, "Accept Changes");

    p = Draw_CachePic ("gfx/bigbox.lmp");
    M_DrawTransPic (160, 64, p);
    p = Draw_CachePic ("gfx/menuplyr.lmp");
    M_BuildTranslationTable (setup_top*16, setup_bottom*16);
    M_DrawTransPicTranslate (180, 72, p);

    M_DrawCharacter (40, setup_cursor_table [setup_cursor], 12+((int)(realtime*4)&1));

    if (setup_cursor == 0)
        M_DrawCharacter (168 + 8*strlen(setup_hostname), setup_cursor_table [setup_cursor], 10+((int)(realtime*4)&1));

    if (setup_cursor == 1)
        M_DrawCharacter (168 + 8*strlen(setup_myname), setup_cursor_table [setup_cursor], 10+((int)(realtime*4)&1));
}


void M_Setup_Key (int k)
{
    int			l;

    switch (k)
    {
    case K_ESCAPE:
        M_Menu_MultiPlayer_f ();
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        setup_cursor--;
        if (setup_cursor < 0)
            setup_cursor = NUM_SETUP_CMDS-1;
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        setup_cursor++;
        if (setup_cursor >= NUM_SETUP_CMDS) setup_cursor = 0;
        break;

    case K_LEFTARROW:
        if (setup_cursor < 2) return;
        S_LocalSound ("misc/menu3.wav");

        if (setup_cursor == 2) setup_top = setup_top - 1;
        if (setup_cursor == 3) setup_bottom = setup_bottom - 1;

        // mh - set the colour
        if (setup_top != setup_oldtop || setup_bottom != setup_oldbottom)
            Cbuf_AddText(va ("color %i %i\n", setup_top, setup_bottom));
        break;

    case K_RIGHTARROW:
        if (setup_cursor < 2) return;
forward:
        S_LocalSound ("misc/menu3.wav");

        if (setup_cursor == 2) setup_top = setup_top + 1;
        if (setup_cursor == 3) setup_bottom = setup_bottom + 1;

        // mh - set the colour
        if (setup_top != setup_oldtop || setup_bottom != setup_oldbottom)
            Cbuf_AddText(va ("color %i %i\n", setup_top, setup_bottom));
        break;

    case K_ENTER:
        if (setup_cursor == 0 || setup_cursor == 1) return;
        if (setup_cursor == 2 || setup_cursor == 3) goto forward;

        // setup_cursor == 4 (OK)
        if (strcmp(cl_name.string, setup_myname) != 0)
            Cbuf_AddText (va ("name \"%s\"\n", setup_myname));

        if (strcmp(hostname.string, setup_hostname) != 0)
            Cvar_Set("hostname", setup_hostname);

        if (setup_top != setup_oldtop || setup_bottom != setup_oldbottom)
            Cbuf_AddText(va ("color %i %i\n", setup_top, setup_bottom));

        m_entersound = true;
        M_Menu_MultiPlayer_f ();
        break;

    case K_BACKSPACE:
        if (setup_cursor == 0)
        {
            if (strlen(setup_hostname))
                setup_hostname[strlen(setup_hostname)-1] = 0;
        }

        if (setup_cursor == 1)
        {
            if (strlen(setup_myname))
                setup_myname[strlen(setup_myname)-1] = 0;
        }
        break;

    default:
        if (k < 32 || k > 127)
            break;

        if (setup_cursor == 0)
        {
            l = strlen(setup_hostname);
            if (l < 15)
            {
                setup_hostname[l+1] = 0;
                setup_hostname[l] = k;
            }
        }
        if (setup_cursor == 1)
        {
            l = strlen(setup_myname);
            if (l < 15)
            {
                setup_myname[l+1] = 0;
                setup_myname[l] = k;
            }
        }
    }

    if (setup_top > 13)
        setup_top = 0;

    if (setup_top < 0)
        setup_top = 13;

    if (setup_bottom > 13)
        setup_bottom = 0;

    if (setup_bottom < 0)
        setup_bottom = 13;
}

//=============================================================================
/* NET MENU */

int	m_net_cursor;
int m_net_items;
int m_net_saveHeight;

char *net_helpMessage [] =
{
    /* .........1.........2.... */
    "                        ",
    " Two computers connected",
    "   through two modems.  ",
    "                        ",

    "                        ",
    " Two computers connected",
    " by a null-modem cable. ",
    "                        ",

    " Novell network LANs    ",
    " or Windows 95 DOS-box. ",
    "                        ",
    "(LAN=Local Area Network)",

    " Commonly used to play  ",
    " over the Internet, but ",
    " also used on a Local   ",
    " Area Network.          "
};

void M_Menu_Net_f (void)
{
    key_dest = key_menu;
    m_state = m_net;
    m_entersound = true;
    m_net_items = 4;

    if (m_net_cursor >= m_net_items)
        m_net_cursor = 0;

    m_net_cursor--;
    M_Net_Key (K_DOWNARROW);
}


void M_Net_Draw (void)
{
    int		f;
    qpic_t	*p;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    f = 32;

    if (serialAvailable)
        p = Draw_CachePic ("gfx/netmen1.lmp");
    else p = NULL;

    if (p) M_DrawTransPic (72, f, p);

    f += 19;

    if (serialAvailable)
        p = Draw_CachePic ("gfx/netmen2.lmp");
    else p = NULL;

    if (p) M_DrawTransPic (72, f, p);

    f += 19;

    if (ipxAvailable)
        p = Draw_CachePic ("gfx/netmen3.lmp");
    else p = Draw_CachePic ("gfx/dim_ipx.lmp");

    M_DrawTransPic (72, f, p);

    f += 19;

    if (tcpipAvailable)
        p = Draw_CachePic ("gfx/netmen4.lmp");
    else p = Draw_CachePic ("gfx/dim_tcp.lmp");

    M_DrawTransPic (72, f, p);

    if (m_net_items == 5)	// JDC, could just be removed
    {
        f += 19;
        p = Draw_CachePic ("gfx/netmen5.lmp");
        M_DrawTransPic (72, f, p);
    }

    f = (320-26*8)/2;
    M_DrawTextBox (f, 134, 24, 4);

    f += 8;
    f -= 24;	// line up the messages properly

    M_Print (f, 142, net_helpMessage[m_net_cursor*4+0]);
    M_Print (f, 150, net_helpMessage[m_net_cursor*4+1]);
    M_Print (f, 158, net_helpMessage[m_net_cursor*4+2]);
    M_Print (f, 166, net_helpMessage[m_net_cursor*4+3]);

    f = (int)(host_time * 10)%6;
    M_DrawTransPic (46, 32 + m_net_cursor * 20, Draw_CachePic (va ("gfx/menudot%i.lmp", f+1)));
}


void M_Net_Key (int k)
{
again:
    switch (k)
    {
    case K_ESCAPE:
        M_Menu_MultiPlayer_f ();
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        if (++m_net_cursor >= m_net_items)
            m_net_cursor = 0;
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        if (--m_net_cursor < 0)
            m_net_cursor = m_net_items - 1;
        break;

    case K_ENTER:
        m_entersound = true;

        switch (m_net_cursor)
        {
        case 0:
            M_Menu_SerialConfig_f ();
            break;

        case 1:
            M_Menu_SerialConfig_f ();
            break;

        case 2:
            M_Menu_LanConfig_f ();
            break;

        case 3:
            M_Menu_LanConfig_f ();
            break;

        case 4:
// multiprotocol
            break;
        }
    }

    if (m_net_cursor == 0 && !serialAvailable)
        goto again;

    if (m_net_cursor == 1 && !serialAvailable)
        goto again;

    if (m_net_cursor == 2 && !ipxAvailable)
        goto again;

    if (m_net_cursor == 3 && !tcpipAvailable)
        goto again;
}

//=============================================================================
/* OPTIONS MENU */

#define	OPTIONS_ITEMS	20
#define	SLIDER_RANGE	10
int submenu_items;

int	options_cursor;
int	g_options_cursor;
int	b_options_cursor;
int	s_options_cursor;
int	d_options_cursor;
int	c_options_cursor;
int	h_options_cursor;
int	f_options_cursor;

// external cvars for options menus
extern cvar_t r_detailsky;
extern cvar_t r_dynamic;
extern cvar_t r_shadows;
extern cvar_t r_waterwave;
extern cvar_t r_particles;
extern cvar_t r_autowatertrans;

extern cvar_t gl_colouredlight;
extern cvar_t gl_vertexlight;
extern cvar_t gl_aliasglow;
extern cvar_t gl_aliastrans;
extern cvar_t gl_fullbright;
extern cvar_t gl_fullbright_alias;
extern cvar_t gl_aliaslightlerp;

extern cvar_t r_interpolate_alias;
extern cvar_t r_interpolate_brush;

extern cvar_t gl_keeptjunctions;
extern cvar_t gl_dlradmult;
extern cvar_t gl_fastbutugly;

extern cvar_t r_bobbing;
extern cvar_t r_muzzleflash;
extern cvar_t r_extradlights;

extern cvar_t gl_fogenable;
extern cvar_t gl_fogstart;
extern cvar_t gl_fogend;
extern cvar_t gl_fogred;
extern cvar_t gl_foggreen;
extern cvar_t gl_fogblue;
extern cvar_t gl_skyfog;

#ifdef USE_MIDI
extern cvar_t midi_audio;
extern cvar_t midi_volumepercent;
#endif

extern cvar_t r_watershade;
extern cvar_t r_waterdlight;
extern cvar_t r_mirroralpha;
extern cvar_t r_torchsmoke;
extern cvar_t r_wateralpha;
extern cvar_t r_novis;

extern cvar_t view_hand;
extern cvar_t temp1;

extern cvar_t cl_mlook;
extern cvar_t cl_fpscount;

extern cvar_t scr_fov;
extern cvar_t v_gamma;


#ifdef USE_MIDI
void MIDI_ChangeVolume (void);
#endif

void M_Menu_Options_f (void)
{
    key_dest = key_menu;
    m_state = m_options;
    m_entersound = true;
}

void M_Menu_General_f (void)
{
    key_dest = key_menu;
    m_state = m_options_general;
    m_entersound = true;
}

void M_Menu_Basic_f (void)
{
    key_dest = key_menu;
    m_state = m_options_basic;
    m_entersound = true;
}

void M_Menu_Static_f (void)
{
    key_dest = key_menu;
    m_state = m_options_static;
    m_entersound = true;
}

void M_Menu_Dynamic_f (void)
{
    key_dest = key_menu;
    m_state = m_options_dynamic;
    m_entersound = true;
}

void M_Menu_Candy_f (void)
{
    key_dest = key_menu;
    m_state = m_options_candy;
    m_entersound = true;
}

void M_Menu_HardCore_f (void)
{
    key_dest = key_menu;
    m_state = m_options_hardcore;
    m_entersound = true;
}

void M_Menu_Fog_f (void)
{
    key_dest = key_menu;
    m_state = m_options_fog;
    m_entersound = true;
}


void M_DrawSlider (int x, int y, float range)
{
    int	i;

    if (range < 0)
        range = 0;

    if (range > 1)
        range = 1;

    x += 8;

    M_DrawCharacter (x-8, y, 128);

    for (i=0 ; i<SLIDER_RANGE ; i++)
        M_DrawCharacter (x + i*8, y, 129);

    M_DrawCharacter (x+i*8, y, 130);
    M_DrawCharacter (x + (SLIDER_RANGE-1)*8 * range, y, 131);
}

void M_DrawCheckbox (int x, int y, int on, char *ontext, char *offtext)
{
    if (on)
        M_Print (x, y, ontext);
    else
        M_Print (x, y, offtext);
}


// a checkbox with 3 values - 0, 1 or 2 - out of bounds values use the 0 text
void M_DrawTripleCheckbox (int x, int y, int test, char *text_0, char *text_1, char *text_2)
{
    if (test == 2)
        M_Print (x, y, text_2);
    else if (test == 1)
        M_Print (x, y, text_1);
    else M_Print (x, y, text_0);
}


// as an alternative to sliders, which may provide us with more useful info about certain
// cvars, this draws a number representing their current value
void M_DrawNumericScale (int x, int y, float num)
{
    char number[8];

    // the horror! the horror!
    if (num > 999.99999)
        sprintf (number, "%0.0f", num);
    else if (num > 99.99999)	// ugh!
        sprintf (number, "%0.0f", num);
    else if (num > 9.99999)
        sprintf (number, "%0.1f", num);
    else sprintf (number, "%0.2f", num);

    M_Print (x, y, number);
}


int current_submenu_cursor;

void M_Options_SubMenu_Init (void)
{
    // draw the options menu - this menu will be under it
    M_Options_Draw (0, &options_store[0]);

    // initialize the cursor position and the number of items - add 8 to the cursor position
    // for each item you add to the main options menu.
    current_submenu_cursor = 136;
    submenu_items = 0;
}


void M_Options_SubMenu_Finish (int menu_cursor, menu_store_t *store)
{
    // don't draw the cursor if there are no options available
    if (submenu_items == 0) return;

    // draw the cursor and do a map reload warning if required
    M_DrawCharacter (208, 136 + menu_cursor * 8, 12 + ((int) (realtime * 4) & 1));

    if (store[menu_cursor].requires_reload == restart_quake)
        M_Print (16, 264, "You Must Restart Quake");
    else if (store[menu_cursor].requires_reload == reload_map)
        M_Print (16, 264, "   Requires Map Reload");
}


void SubMenu_AdjustSliders (int dir, menu_store_t *store, int store_num)
{
    switch (store[store_num].item_type)
    {
    case on_off:
        Cvar_SetValue (store[store_num].cvar_store->name, !store[store_num].cvar_store->value);
        break;

    case on_off2:
        if (store[store_num].cvar_store->value > store[store_num].min_scale)
            store[store_num].cvar_store->value = store[store_num].min_scale;
        else
            store[store_num].cvar_store->value = store[store_num].max_scale;

        Cvar_SetValue (store[store_num].cvar_store->name, store[store_num].cvar_store->value);
        break;

    case three_item:
        // much the same as below only it wraps rather than stopping at min/max,
        store[store_num].cvar_store->value += (float) dir;

        if (store[store_num].cvar_store->value > 2.0)
            store[store_num].cvar_store->value = 0;

        if (store[store_num].cvar_store->value < 0)
            store[store_num].cvar_store->value = 2.0;

        Cvar_SetValue (store[store_num].cvar_store->name, store[store_num].cvar_store->value);
        break;

    case slider:
    case scale:
        // these increment and decrement the same, they just draw differently
        store[store_num].cvar_store->value += (store[store_num].slider_step * (float) dir);

        if (store[store_num].cvar_store->value > store[store_num].max_scale)
            store[store_num].cvar_store->value = store[store_num].max_scale;

        if (store[store_num].cvar_store->value < store[store_num].min_scale)
            store[store_num].cvar_store->value = store[store_num].min_scale;

        Cvar_SetValue (store[store_num].cvar_store->name, store[store_num].cvar_store->value);

        // ugly hack - for midi sound, we need to adjust the volume right now...
#ifdef USE_MIDI
        if (!strcmp (store[store_num].cvar_store->name, "midi_volumepercent")) MIDI_ChangeVolume ();
#endif
        break;

    case sub_menu:
        // a pointer to a function within a structure - woo-hoo!  this is basically a c++
        // class member type thing coded in standard c - so there!
        (*store[store_num].submenu_func) ();
        break;

    case spacer:
        // literally does nothing!
        // this condition should never happen, so if it does we can add something in here
        // to let us know that something has gone wrong.
        break;

    default:
        // should never happen, but just in case...
        break;
    }
}


// check for a spacer if we move up or down and if we find one, go to the next
// option up or down
int Check_Spacers (int dir, int submenu_options_cursor, menu_store_t *store)
{
    int overload;

    S_LocalSound ("misc/menu1.wav");

    overload = 0;

    while (1)
    {
        overload++;

        submenu_options_cursor += dir;
        if (submenu_options_cursor < 0) submenu_options_cursor = submenu_items - 1;
        if (submenu_options_cursor >= submenu_items) submenu_options_cursor = 0;
        if (store[submenu_options_cursor].item_type != spacer) return submenu_options_cursor;

        // if we've gone through every option and can't find any that ain't a spacer,
        // something's gone seriously wrong, so get out!
        if (overload > submenu_items + 1) return -1;
    }
}

int M_SubMenu_Key (int k, int submenu_options_cursor, menu_store_t *store)
{
    // if there's nothing in the submenu, get out
    if (submenu_items == 0)
    {
        (*Previous_Options_Menu) ();
        return submenu_options_cursor;
    }

    // to do: check for spacer types when arrowing up or down...
    switch (k)
    {
    case K_ESCAPE:
        (*Previous_Options_Menu) ();
        break;

    case K_ENTER:
        m_entersound = true;
        SubMenu_AdjustSliders (1, store, submenu_options_cursor);
        return submenu_options_cursor;

    case K_UPARROW:
        submenu_options_cursor = Check_Spacers (-1, submenu_options_cursor, store);

        // error checking
        if (submenu_options_cursor == -1)
        {
            M_Menu_Main_f ();
            return 0;
        }
        break;

    case K_DOWNARROW:
        submenu_options_cursor = Check_Spacers (1, submenu_options_cursor, store);

        // error checking
        if (submenu_options_cursor == -1)
        {
            M_Menu_Main_f ();
            return 0;
        }
        break;

    case K_LEFTARROW:
        SubMenu_AdjustSliders (-1, store, submenu_options_cursor);
        break;

    case K_RIGHTARROW:
        SubMenu_AdjustSliders (1, store, submenu_options_cursor);
        break;

    default:
        break;
    }

    return submenu_options_cursor;
}


void M_General_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();
    M_MenuItem_Draw (1.0 - (v_gamma.value / 2.0), slider, "Brightness", "MH", "MH", "MH", &v_gamma, 0.0, 2.0, current_store, no_reload, -0.025, M_FuncNull); // lower is brighter
    M_MenuItem_Draw (r_detailsky.value, on_off, "Detailed Sky", "On", "Off", "MH", &r_detailsky, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (cl_fpscount.value, on_off, "FPS Counter", "On", "Off", "MH", &cl_fpscount, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (view_hand.value, three_item, "Viewmodel", "Center", "Right", "Left", &view_hand, 0.0, 2.0, current_store, no_reload, 1.0, M_FuncNull);
    M_MenuItem_Draw (r_watershade.value * 1.25, slider, "Water Shading", "MH", "MH", "MH", &r_watershade, 0.0, 0.8, current_store, no_reload, 0.05, M_FuncNull);
    M_MenuItem_Draw (((scr_fov.value - 10) / 160), slider, "Field of View", "MH", "MH", "MH", &scr_fov, 10.0, 170.0, current_store, no_reload, 5.0, M_FuncNull);
    M_MenuItem_Draw (r_bobbing.value, on_off, "Bobbing Pickups", "On", "Off", "MH", &r_bobbing, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_waterwarp.value, on_off, "Water Warping", "Realistic", "Simple", "MH", &r_waterwarp, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_skyfog.value, on_off, "Sky Fogging", "On", "Off", "MH", &gl_skyfog, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_Options_SubMenu_Finish (g_options_cursor, &general_store[0]);
}


void M_Basic_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();
    M_MenuItem_Draw (gl_fastbutugly.value, on_off, "Fast but Ugly", "On", "Off", "MH", &gl_fastbutugly, 0.0, 0.0, current_store, reload_map, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_keeptjunctions.value, on_off, "Keep T Junctions", "Slow", "Fast", "MH", &gl_keeptjunctions, 0.0, 0.0, current_store, reload_map, 0.0, M_FuncNull);
    M_Options_SubMenu_Finish (b_options_cursor, &basic_store[0]);
}


void M_Static_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();
    M_MenuItem_Draw (gl_colouredlight.value, on_off, "Coloured Light", "On", "Off", "MH", &gl_colouredlight, 0.0, 0.0, current_store, true, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_vertexlight.value, on_off, "Vertex Lighting", "On", "Off", "MH", &gl_vertexlight, 0.0, 0.0, current_store, false, 0.0, M_FuncNull);
    M_Options_SubMenu_Finish (s_options_cursor, &static_store[0]);
}


void M_Dynamic_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();
    M_MenuItem_Draw (r_dynamic.value, on_off, "Dynamic Light", "On", "Off", "MH", &r_dynamic, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_dlradmult.value, scale, "Dynamic Light Radius", "MH", "MH", "MH", &gl_dlradmult, 0.5, 1.5, current_store, no_reload, 0.05, M_FuncNull);
    M_MenuItem_Draw (r_muzzleflash.value, on_off, "Muzzle Flashes", "On", "Off", "MH", &r_muzzleflash, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_extradlights.value, on_off, "Extra Dynamic Lights", "On", "Off", "MH", &r_extradlights, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_waterdlight.value, on_off, "Water Dynamic Lights", "On", "Off", "MH", &r_waterdlight, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_Options_SubMenu_Finish (d_options_cursor, &dynamic_store[0]);
}


void M_Candy_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();
    M_MenuItem_Draw (r_interpolate_alias.value, on_off, "Alias Interpolation", "On", "Off", "MH", &r_interpolate_alias, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_interpolate_brush.value, on_off, "Brush Interpolation", "On", "Off", "MH", &r_interpolate_brush, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_fullbright.value, on_off, "Fullbright Tex Colours", "On", "Off", "MH", &gl_fullbright, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_fullbright_alias.value, on_off, "Fullbright MDL Colours", "On", "Off", "MH", &gl_fullbright_alias, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_aliasglow.value, on_off, "Glowing Alias Models", "On", "Off", "MH", &gl_aliasglow, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_aliastrans.value, on_off, "Translucent Models", "On", "Off", "MH", &gl_aliastrans, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (gl_aliaslightlerp.value, on_off, "Model Light Lerping", "On", "Off", "MH", &gl_aliaslightlerp, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_particles.value, on_off, "Particle System", "Enhanced", "Classic", "MH", &r_particles, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_torchsmoke.value, on_off, "Smoking Torches", "On", "Off", "MH", &r_torchsmoke, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_autowatertrans.value, on_off, "Auto Trans Water", "On", "Off", "MH", &r_autowatertrans, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw ((1.0 - r_wateralpha.value), slider, "Water Translucency", "MH", "MH", "MH", &r_wateralpha, 0.0, 1.0, current_store, no_reload, -0.05, M_FuncNull);
    M_Options_SubMenu_Finish (c_options_cursor, &candy_store[0]);
}


void M_HardCore_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();
    M_MenuItem_Draw ((1.0 - r_mirroralpha.value), slider, "Mirror Reflectivity", "MH", "MH", "MH", &r_mirroralpha, 0.0, 1.0, current_store, no_reload, -0.05, M_FuncNull);
    M_MenuItem_Draw (r_shadows.value, on_off, "Shadows", "On", "Off", "MH", &r_shadows, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (r_novis.value, on_off, "Use Visibility Info", "No", "Yes", "MH", &r_novis, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw ((r_waterwave.value / 3.0), slider, "Water Waves", "MH", "MH", "MH", &r_waterwave, 0.0, 3.0, current_store, no_reload, 3.0, M_FuncNull);
    M_Options_SubMenu_Finish (h_options_cursor, &hardcore_store[0]);
}


void M_Fog_Draw (menu_store_t *current_store)
{
    M_Options_SubMenu_Init ();

    M_MenuItem_Draw (gl_fogenable.value, on_off, "Global Fog", "On", "Off", "MH", &gl_fogenable, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);

    if (gl_fogenable.value)
    {
        // prevent any silliness with gl_fogstart and gl_fogend
        if (gl_fogstart.value >= gl_fogend.value) Cvar_SetValue ("gl_fogstart", gl_fogend.value - 1);
        if (gl_fogend.value <= gl_fogstart.value) Cvar_SetValue ("gl_fogend", gl_fogstart.value + 1);

        M_MenuItem_Draw ((gl_fogstart.value / 4096), slider, "Fog Start", "MH", "MH", "MH", &gl_fogstart, 5.0, 4095.0, current_store, no_reload, 102.4, M_FuncNull);
        M_MenuItem_Draw ((gl_fogend.value / 4096), slider, "Fog End", "MH", "MH", "MH", &gl_fogend, 6.0, 4096.0, current_store, no_reload, 102.4, M_FuncNull);
        M_MenuItem_Draw (gl_fogred.value, slider, "Fog Red", "MH", "MH", "MH", &gl_fogred, 0.0, 1.0, current_store, no_reload, 0.1, M_FuncNull);
        M_MenuItem_Draw (gl_foggreen.value, slider, "Fog Green", "MH", "MH", "MH", &gl_foggreen, 0.0, 1.0, current_store, no_reload, 0.1, M_FuncNull);
        M_MenuItem_Draw (gl_fogblue.value, slider, "Fog Blue", "MH", "MH", "MH", &gl_fogblue, 0.0, 1.0, current_store, no_reload, 0.1, M_FuncNull);
    }
    else
    {
        // in case gl_fogenable was set to 0 from the console...
        f_options_cursor = 0;
    }

    M_Options_SubMenu_Finish (f_options_cursor, &fog_store[0]);
}


void M_Menu_GoToConsole_f (void)
{
    m_state = m_none;
    Con_ToggleConsole_f ();
}


void M_Menu_ResetToDefaults_f (void)
{
    Cbuf_AddText ("exec default.cfg\n");
}


void M_Options_Draw (int draw_cursor, menu_store_t *current_store)
{
    qpic_t	*p;

    p = Draw_CachePic ("gfx/p_option.lmp");
    M_DrawPic ((320-p->width)/2, 4, p);

    // init these seperately rather than passing them through the main init function cos
    // that would give us an evil infinite loop, and besides, the cursor position in
    // that function is incorrect for this menu
    current_submenu_cursor = 32;
    submenu_items = 0;

    // print the menu
    M_MenuItem_Draw (temp1.value, sub_menu, "Customize Controls", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_Keys_f);
    M_MenuItem_Draw (temp1.value, sub_menu, "Go to Console", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_GoToConsole_f);
    M_MenuItem_Draw (temp1.value, sub_menu, "Reset to Defaults", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_ResetToDefaults_f);
    M_MenuItem_Draw (temp1.value, spacer, " ", " ", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (((sensitivity.value - 1) / 10), slider, "Mouse Speed", "MH", "MH", "MH", &sensitivity, 1.0, 11.0, current_store, no_reload, 0.5, M_FuncNull);
    M_MenuItem_Draw ((float) (m_pitch.value < 0), on_off2, "Invert Mouse", "On", "Off", "MH", &m_pitch, -0.022, 0.022, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (cl_mlook.value, on_off, "Mouse Look", "Free", "Fixed", "MH", &cl_mlook, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw ((float) (cl_forwardspeed.value > 200), on_off2, "Always Run", "On", "Off", "MH", &cl_forwardspeed, 200.0, 400.0, current_store, no_reload, 0.0, M_FuncNull);
    M_MenuItem_Draw (temp1.value, spacer, " ", " ", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
#ifdef USE_MIDI
    M_MenuItem_Draw (midi_audio.value, on_off, "Music Type", "MIDI", "CD", "MH", &midi_audio, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
#else
    M_MenuItem_Draw (temp1.value, spacer, "Music Type", "CD Only", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);
#endif

#ifdef USE_MIDI
    if (midi_audio.value)
        M_MenuItem_Draw ((midi_volumepercent.value / 6553.5), scale, "MIDI Volume", "MH", "MH", "MH", &midi_volumepercent, 0.0, 65535.0, current_store, no_reload, 3276.75, M_FuncNull);
    else
#endif
        M_MenuItem_Draw (bgmvolume.value, scale, "CD Volume", "MH", "MH", "MH", &bgmvolume, 0.0, 1.0, current_store, no_reload, 1.0, M_FuncNull);

    M_MenuItem_Draw (volume.value, scale, "Effects Volume", "MH", "MH", "MH", &volume, 0.0, 1.0, current_store, no_reload, 0.1, M_FuncNull);
    M_MenuItem_Draw (temp1.value, spacer, " ", " ", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_FuncNull);

    if (m_state == m_options)
    {
        // only draw the submenus if we're in the main options menu
        M_MenuItem_Draw (temp1.value, sub_menu, "General View Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_General_f);
        M_MenuItem_Draw (temp1.value, sub_menu, "Basic Speed Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_Basic_f);
        M_MenuItem_Draw (temp1.value, sub_menu, "Static Light Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_Static_f);
        M_MenuItem_Draw (temp1.value, sub_menu, "Dynamic Light Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_Dynamic_f);
        M_MenuItem_Draw (temp1.value, sub_menu, "Basic F/X Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_Candy_f);
        M_MenuItem_Draw (temp1.value, sub_menu, "Hardcore F/X Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_HardCore_f);
        M_MenuItem_Draw (temp1.value, sub_menu, "Global Fog Options", "MH", "MH", "MH", &temp1, 0.0, 0.0, current_store, no_reload, 0.0, M_Menu_Fog_f);

        Previous_Options_Menu = M_Menu_Main_f;
    }
    else Previous_Options_Menu = M_Menu_Options_f;

    // cursor
    if (draw_cursor) M_DrawCharacter (208, 32 + options_cursor*8, 12+((int)(realtime*4)&1));

    if (current_store[options_cursor].requires_reload == restart_quake)
        M_Print (16, 264, "You Must Restart Quake");
    else if (current_store[options_cursor].requires_reload == reload_map)
        M_Print (16, 264, "   Requires Map Reload");
}


//=============================================================================
/* KEYS MENU */

char *bindnames[][2] =
{
    {"+attack", 		"Attack"},
    {"impulse 10", 		"Next Weapon"},
    {"+jump", 			"Jump/Swim Up"},
    {"+forward", 		"Walk Forward"},
    {"+back", 			"Backpedal"},
    {"+left", 			"Turn Left"},
    {"+right", 			"Turn Right"},
    {"+speed", 			"Run"},
    {"+moveleft", 		"Step Left"},
    {"+moveright", 		"Step Right"},
    {"+strafe", 		"Sidestep"},
    {"+lookup", 		"Look Up"},
    {"+lookdown", 		"Look Down"},
    {"centerview", 		"Center View"},
    {"+mlook", 			"Mouse Look"},
    {"+klook", 			"Keyboard Look"},
    {"+moveup",			"Swim Up"},
    {"+movedown",		"Swim Down"}
};

#define	NUMCOMMANDS	(sizeof(bindnames)/sizeof(bindnames[0]))

int		keys_cursor;
int		bind_grab;

void M_Menu_Keys_f (void)
{
    key_dest = key_menu;
    m_state = m_keys;
    m_entersound = true;
}


void M_FindKeysForCommand (char *command, int *twokeys)
{
    int		count;
    int		j;
    int		l;
    char	*b;

    twokeys[0] = twokeys[1] = -1;
    l = strlen(command);
    count = 0;

    for (j=0 ; j<256 ; j++)
    {
        b = keybindings[j];

        if (!b)
            continue;

        if (!strncmp (b, command, l) )
        {
            twokeys[count] = j;
            count++;

            if (count == 2)
                break;
        }
    }
}


void M_UnbindCommand (char *command)
{
    int		j;
    int		l;
    char	*b;

    l = strlen(command);

    for (j=0 ; j<256 ; j++)
    {
        b = keybindings[j];

        if (!b)
            continue;

        if (!strncmp (b, command, l) )
            Key_SetBinding (j, "");
    }
}


void M_Keys_Draw (void)
{
    int		i, l;
    int		keys[2];
    char	*name;
    int		x, y;
    qpic_t	*p;

    p = Draw_CachePic ("gfx/ttl_cstm.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    if (bind_grab)
        M_Print (12, 32, "Press a key or button for this action");
    else
        M_Print (18, 32, "Enter to change, backspace to clear");

    // search for known bindings
    for (i=0 ; i<NUMCOMMANDS ; i++)
    {
        y = 48 + 8*i;

        M_Print (16, y, bindnames[i][1]);

        l = strlen (bindnames[i][0]);

        M_FindKeysForCommand (bindnames[i][0], keys);

        if (keys[0] == -1)
        {
            M_Print (140, y, "???");
        }
        else
        {
            name = Key_KeynumToString (keys[0]);
            M_Print (140, y, name);
            x = strlen(name) * 8;
            if (keys[1] != -1)
            {
                M_Print (140 + x + 8, y, "or");
                M_Print (140 + x + 32, y, Key_KeynumToString (keys[1]));
            }
        }
    }

    if (bind_grab)
        M_DrawCharacter (138, 48 + keys_cursor*8, '=');
    else
        M_DrawCharacter (138, 48 + keys_cursor*8, 12+((int)(realtime*4)&1));
}


void M_Keys_Key (int k)
{
    char	cmd[80];
    int		keys[2];

    if (bind_grab)
    {
        // defining a key
        S_LocalSound ("misc/menu1.wav");

        if (k == K_ESCAPE)
        {
            bind_grab = false;
        }
        else if (k != '`')
        {
            sprintf (cmd, "bind \"%s\" \"%s\"\n", Key_KeynumToString (k), bindnames[keys_cursor][0]);
            Cbuf_InsertText (cmd);
        }

        bind_grab = false;
        return;
    }

    switch (k)
    {
    case K_ESCAPE:
        M_Menu_Options_f ();
        break;

    case K_LEFTARROW:
    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        keys_cursor--;
        if (keys_cursor < 0)
            keys_cursor = NUMCOMMANDS-1;
        break;

    case K_DOWNARROW:
    case K_RIGHTARROW:
        S_LocalSound ("misc/menu1.wav");
        keys_cursor++;
        if (keys_cursor >= NUMCOMMANDS)
            keys_cursor = 0;
        break;

    case K_ENTER:		// go into bind mode
        M_FindKeysForCommand (bindnames[keys_cursor][0], keys);
        S_LocalSound ("misc/menu2.wav");
        if (keys[1] != -1)
            M_UnbindCommand (bindnames[keys_cursor][0]);
        bind_grab = true;
        break;

    case K_BACKSPACE:		// delete bindings
    case K_DEL:				// delete bindings
        S_LocalSound ("misc/menu2.wav");
        M_UnbindCommand (bindnames[keys_cursor][0]);
        break;
    }
}

//=============================================================================
/* HELP MENU */

int		help_page;
#if 0
#define	NUM_HELP_PAGES	8
#else
#define	NUM_HELP_PAGES	6
#endif


void M_Menu_Help_f (void)
{
    key_dest = key_menu;
    m_state = m_help;
    m_entersound = true;
    help_page = 0;
}



void M_Help_Draw (void)
{
    M_DrawPic (0, 8, Draw_CachePic (va ("gfx/help%i.lmp", help_page)));
}


void M_Help_Key (int key)
{
    switch (key)
    {
    case K_ESCAPE:
        M_Menu_Main_f ();
        break;

    case K_UPARROW:
    case K_RIGHTARROW:
        m_entersound = true;
        if (++help_page >= NUM_HELP_PAGES)
            help_page = NUM_HELP_PAGES - 1;
        break;

    case K_DOWNARROW:
    case K_LEFTARROW:
        m_entersound = true;
        if (--help_page < 0)
            help_page = 0;
        break;
    }

}

//=============================================================================
/* QUIT MENU */

int		msgNumber;
int		m_quit_prevstate;
qboolean	wasInMenus;


void M_Menu_Quit_f (void)
{
    if (m_state == m_quit)
        return;

    wasInMenus = (key_dest == key_menu);
    key_dest = key_menu;
    m_quit_prevstate = m_state;
    m_state = m_quit;
    m_entersound = true;
    msgNumber = rand()&7;
}


void M_Quit_Key (int key)
{
    switch (key)
    {
    case K_ESCAPE:
    case 'n':
    case 'N':
        if (wasInMenus)
        {
            m_state = m_quit_prevstate;
            m_entersound = true;
        }
        else
        {
            key_dest = key_game;
            m_state = m_none;
        }
        break;

    case 'Y':
    case 'y':
        key_dest = key_console;
        Host_Quit_f ();
        break;

    default:
        break;
    }

}


void M_Quit_Draw (void)
{
    if (wasInMenus)
    {
        m_state = m_quit_prevstate;
        m_recursiveDraw = false;
        M_Draw ();
        m_state = m_quit;
    }

    M_DrawTextBox (0, 0, 40, 34);
    M_PrintWhite	(4, 20,  "-- You have been watching MHQUAKE! --");

    M_Print			(4, 36,  "A total Audio/Visual/Neural sensory  ");
    M_Print			(4, 44,  "overload experience, brought to you  ");
    M_Print			(4, 52,  "by those nice people at the          ");
    M_Print			(4, 60,  "Amarillo Hoss Trading Company.       ");

    M_Print			(4, 76,  "For some reason, you seem to have had");
    M_Print			(4, 84,  "enough MHQuake for now. Why this is  ");
    M_Print			(4, 92,  "so is your own business.             ");

    M_Print			(4, 108, "MHQuake does not care if you hate it.");
    M_Print			(4, 116, "MHQuake loves you nonetheless.       ");
    M_Print			(4, 124, "MHQuake loves everybody.");

    M_Print			(4, 140, "MHQuake does not care if you think   ");
    M_Print			(4, 148, "that you've wasted money downloading ");
    M_Print			(4, 156, "it. MHQuake prefers to see you waste ");
    M_Print			(4, 164, "money in this way than in some of the");
    M_Print			(4, 172, "other ways available. At least you   ");
    M_Print			(4, 180, "won't go blind...                    ");

    M_Print			(4, 196, "If you really want to quit now, press");
    M_Print			(4, 204, "Y to do so, otherwise press N for    ");
    M_Print			(4, 212, "more MHQuake fun.                    ");

    M_Print			(4, 228, "MHQuake supports free goats for      ");
    M_Print			(4, 236, "the over 50s.                        ");

    M_PrintWhite	(4, 252, "-- Lay down thy Nailgun and groove --");
}

//=============================================================================

/* SERIAL CONFIG MENU */

int		serialConfig_cursor;
int		serialConfig_cursor_table[] = {48, 64, 80, 96, 112, 132};
#define	NUM_SERIALCONFIG_CMDS	6

static int ISA_uarts[]	= {0x3f8,0x2f8,0x3e8,0x2e8};
static int ISA_IRQs[]	= {4,3,4,3};
int serialConfig_baudrate[] = {9600,14400,19200,28800,38400,57600};

int		serialConfig_comport;
int		serialConfig_irq ;
int		serialConfig_baud;
char	serialConfig_phone[16];

void M_Menu_SerialConfig_f (void)
{
    int		n;
    int		port;
    int		baudrate;
    qboolean	useModem;

    key_dest = key_menu;
    m_state = m_serialconfig;
    m_entersound = true;
    if (JoiningGame && SerialConfig)
        serialConfig_cursor = 4;
    else
        serialConfig_cursor = 5;

    (*GetComPortConfig) (0, &port, &serialConfig_irq, &baudrate, &useModem);

    // map uart's port to COMx
    for (n = 0; n < 4; n++)
        if (ISA_uarts[n] == port)
            break;
    if (n == 4)
    {
        n = 0;
        serialConfig_irq = 4;
    }
    serialConfig_comport = n + 1;

    // map baudrate to index
    for (n = 0; n < 6; n++)
        if (serialConfig_baudrate[n] == baudrate)
            break;
    if (n == 6)
        n = 5;
    serialConfig_baud = n;

    m_return_onerror = false;
    m_return_reason[0] = 0;
}


void M_SerialConfig_Draw (void)
{
    qpic_t	*p;
    int		basex;
    char	*startJoin;
    char	*directModem;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    basex = (320-p->width)/2;
    M_DrawPic (basex, 4, p);

    if (StartingGame)
        startJoin = "New Game";
    else
        startJoin = "Join Game";
    if (SerialConfig)
        directModem = "Modem";
    else
        directModem = "Direct Connect";
    M_Print (basex, 32, va ("%s - %s", startJoin, directModem));
    basex += 8;

    M_Print (basex, serialConfig_cursor_table[0], "Port");
    M_DrawTextBox (160, 40, 4, 1);
    M_Print (168, serialConfig_cursor_table[0], va("COM%u", serialConfig_comport));

    M_Print (basex, serialConfig_cursor_table[1], "IRQ");
    M_DrawTextBox (160, serialConfig_cursor_table[1]-8, 1, 1);
    M_Print (168, serialConfig_cursor_table[1], va("%u", serialConfig_irq));

    M_Print (basex, serialConfig_cursor_table[2], "Baud");
    M_DrawTextBox (160, serialConfig_cursor_table[2]-8, 5, 1);
    M_Print (168, serialConfig_cursor_table[2], va("%u", serialConfig_baudrate[serialConfig_baud]));

    if (SerialConfig)
    {
        M_Print (basex, serialConfig_cursor_table[3], "Modem Setup...");
        if (JoiningGame)
        {
            M_Print (basex, serialConfig_cursor_table[4], "Phone number");
            M_DrawTextBox (160, serialConfig_cursor_table[4]-8, 16, 1);
            M_Print (168, serialConfig_cursor_table[4], serialConfig_phone);
        }
    }

    if (JoiningGame)
    {
        M_DrawTextBox (basex, serialConfig_cursor_table[5]-8, 7, 1);
        M_Print (basex+8, serialConfig_cursor_table[5], "Connect");
    }
    else
    {
        M_DrawTextBox (basex, serialConfig_cursor_table[5]-8, 2, 1);
        M_Print (basex+8, serialConfig_cursor_table[5], "OK");
    }

    M_DrawCharacter (basex-8 + 8, serialConfig_cursor_table [serialConfig_cursor], 12+((int)(realtime*4)&1));

    if (serialConfig_cursor == 4)
        M_DrawCharacter (176 + 8*strlen(serialConfig_phone), serialConfig_cursor_table [serialConfig_cursor], 10+((int)(realtime*4)&1));

    if (*m_return_reason)
        M_PrintWhite (basex, 148, m_return_reason);
}


void M_SerialConfig_Key (int key)
{
    int		l;

    switch (key)
    {
    case K_ESCAPE:
        M_Menu_Net_f ();
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        serialConfig_cursor--;
        if (serialConfig_cursor < 0)
            serialConfig_cursor = NUM_SERIALCONFIG_CMDS-1;
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        serialConfig_cursor++;
        if (serialConfig_cursor >= NUM_SERIALCONFIG_CMDS)
            serialConfig_cursor = 0;
        break;

    case K_LEFTARROW:
        if (serialConfig_cursor > 2)
            break;
        S_LocalSound ("misc/menu3.wav");

        if (serialConfig_cursor == 0)
        {
            serialConfig_comport--;
            if (serialConfig_comport == 0)
                serialConfig_comport = 4;
            serialConfig_irq = ISA_IRQs[serialConfig_comport-1];
        }

        if (serialConfig_cursor == 1)
        {
            serialConfig_irq--;
            if (serialConfig_irq == 6)
                serialConfig_irq = 5;
            if (serialConfig_irq == 1)
                serialConfig_irq = 7;
        }

        if (serialConfig_cursor == 2)
        {
            serialConfig_baud--;
            if (serialConfig_baud < 0)
                serialConfig_baud = 5;
        }

        break;

    case K_RIGHTARROW:
        if (serialConfig_cursor > 2)
            break;
forward:
        S_LocalSound ("misc/menu3.wav");

        if (serialConfig_cursor == 0)
        {
            serialConfig_comport++;
            if (serialConfig_comport > 4)
                serialConfig_comport = 1;
            serialConfig_irq = ISA_IRQs[serialConfig_comport-1];
        }

        if (serialConfig_cursor == 1)
        {
            serialConfig_irq++;
            if (serialConfig_irq == 6)
                serialConfig_irq = 7;
            if (serialConfig_irq == 8)
                serialConfig_irq = 2;
        }

        if (serialConfig_cursor == 2)
        {
            serialConfig_baud++;
            if (serialConfig_baud > 5)
                serialConfig_baud = 0;
        }

        break;

    case K_ENTER:
        if (serialConfig_cursor < 3)
            goto forward;

        m_entersound = true;

        if (serialConfig_cursor == 3)
        {
            (*SetComPortConfig) (0, ISA_uarts[serialConfig_comport-1], serialConfig_irq, serialConfig_baudrate[serialConfig_baud], SerialConfig);

            M_Menu_ModemConfig_f ();
            break;
        }

        if (serialConfig_cursor == 4)
        {
            serialConfig_cursor = 5;
            break;
        }

        // serialConfig_cursor == 5 (OK/CONNECT)
        (*SetComPortConfig) (0, ISA_uarts[serialConfig_comport-1], serialConfig_irq, serialConfig_baudrate[serialConfig_baud], SerialConfig);

        M_ConfigureNetSubsystem ();

        if (StartingGame)
        {
            M_Menu_GameOptions_f ();
            break;
        }

        m_return_state = m_state;
        m_return_onerror = true;
        key_dest = key_game;
        m_state = m_none;

        if (SerialConfig)
            Cbuf_AddText (va ("connect \"%s\"\n", serialConfig_phone));
        else
            Cbuf_AddText ("connect\n");
        break;

    case K_BACKSPACE:
        if (serialConfig_cursor == 4)
        {
            if (strlen(serialConfig_phone))
                serialConfig_phone[strlen(serialConfig_phone)-1] = 0;
        }
        break;

    default:
        if (key < 32 || key > 127)
            break;
        if (serialConfig_cursor == 4)
        {
            l = strlen(serialConfig_phone);
            if (l < 15)
            {
                serialConfig_phone[l+1] = 0;
                serialConfig_phone[l] = key;
            }
        }
    }

    if (DirectConfig && (serialConfig_cursor == 3 || serialConfig_cursor == 4))
        if (key == K_UPARROW)
            serialConfig_cursor = 2;
        else
            serialConfig_cursor = 5;

    if (SerialConfig && StartingGame && serialConfig_cursor == 4)
        if (key == K_UPARROW)
            serialConfig_cursor = 3;
        else
            serialConfig_cursor = 5;
}

//=============================================================================
/* MODEM CONFIG MENU */

int		modemConfig_cursor;
int		modemConfig_cursor_table [] = {40, 56, 88, 120, 156};
#define NUM_MODEMCONFIG_CMDS	5

char	modemConfig_dialing;
char	modemConfig_clear [16];
char	modemConfig_init [32];
char	modemConfig_hangup [16];

void M_Menu_ModemConfig_f (void)
{
    key_dest = key_menu;
    m_state = m_modemconfig;
    m_entersound = true;
    (*GetModemConfig) (0, &modemConfig_dialing, modemConfig_clear, modemConfig_init, modemConfig_hangup);
}


void M_ModemConfig_Draw (void)
{
    qpic_t	*p;
    int		basex;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    basex = (320-p->width)/2;
    M_DrawPic (basex, 4, p);
    basex += 8;

    if (modemConfig_dialing == 'P')
        M_Print (basex, modemConfig_cursor_table[0], "Pulse Dialing");
    else
        M_Print (basex, modemConfig_cursor_table[0], "Touch Tone Dialing");

    M_Print (basex, modemConfig_cursor_table[1], "Clear");
    M_DrawTextBox (basex, modemConfig_cursor_table[1]+4, 16, 1);
    M_Print (basex+8, modemConfig_cursor_table[1]+12, modemConfig_clear);
    if (modemConfig_cursor == 1)
        M_DrawCharacter (basex+16 + 8*strlen(modemConfig_clear), modemConfig_cursor_table[1]+12, 10+((int)(realtime*4)&1));

    M_Print (basex, modemConfig_cursor_table[2], "Init");
    M_DrawTextBox (basex, modemConfig_cursor_table[2]+4, 30, 1);
    M_Print (basex+8, modemConfig_cursor_table[2]+12, modemConfig_init);
    if (modemConfig_cursor == 2)
        M_DrawCharacter (basex+16 + 8*strlen(modemConfig_init), modemConfig_cursor_table[2]+12, 10+((int)(realtime*4)&1));

    M_Print (basex, modemConfig_cursor_table[3], "Hangup");
    M_DrawTextBox (basex, modemConfig_cursor_table[3]+4, 16, 1);
    M_Print (basex+8, modemConfig_cursor_table[3]+12, modemConfig_hangup);
    if (modemConfig_cursor == 3)
        M_DrawCharacter (basex+16 + 8*strlen(modemConfig_hangup), modemConfig_cursor_table[3]+12, 10+((int)(realtime*4)&1));

    M_DrawTextBox (basex, modemConfig_cursor_table[4]-8, 2, 1);
    M_Print (basex+8, modemConfig_cursor_table[4], "OK");

    M_DrawCharacter (basex, modemConfig_cursor_table [modemConfig_cursor], 12+((int)(realtime*4)&1));
}


void M_ModemConfig_Key (int key)
{
    int		l;

    switch (key)
    {
    case K_ESCAPE:
        M_Menu_SerialConfig_f ();
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        modemConfig_cursor--;
        if (modemConfig_cursor < 0)
            modemConfig_cursor = NUM_MODEMCONFIG_CMDS-1;
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        modemConfig_cursor++;
        if (modemConfig_cursor >= NUM_MODEMCONFIG_CMDS)
            modemConfig_cursor = 0;
        break;

    case K_LEFTARROW:
    case K_RIGHTARROW:
        if (modemConfig_cursor == 0)
        {
            if (modemConfig_dialing == 'P')
                modemConfig_dialing = 'T';
            else
                modemConfig_dialing = 'P';
            S_LocalSound ("misc/menu1.wav");
        }
        break;

    case K_ENTER:
        if (modemConfig_cursor == 0)
        {
            if (modemConfig_dialing == 'P')
                modemConfig_dialing = 'T';
            else
                modemConfig_dialing = 'P';
            m_entersound = true;
        }

        if (modemConfig_cursor == 4)
        {
            (*SetModemConfig) (0, va ("%c", modemConfig_dialing), modemConfig_clear, modemConfig_init, modemConfig_hangup);
            m_entersound = true;
            M_Menu_SerialConfig_f ();
        }
        break;

    case K_BACKSPACE:
        if (modemConfig_cursor == 1)
        {
            if (strlen(modemConfig_clear))
                modemConfig_clear[strlen(modemConfig_clear)-1] = 0;
        }

        if (modemConfig_cursor == 2)
        {
            if (strlen(modemConfig_init))
                modemConfig_init[strlen(modemConfig_init)-1] = 0;
        }

        if (modemConfig_cursor == 3)
        {
            if (strlen(modemConfig_hangup))
                modemConfig_hangup[strlen(modemConfig_hangup)-1] = 0;
        }
        break;

    default:
        if (key < 32 || key > 127)
            break;

        if (modemConfig_cursor == 1)
        {
            l = strlen(modemConfig_clear);
            if (l < 15)
            {
                modemConfig_clear[l+1] = 0;
                modemConfig_clear[l] = key;
            }
        }

        if (modemConfig_cursor == 2)
        {
            l = strlen(modemConfig_init);
            if (l < 29)
            {
                modemConfig_init[l+1] = 0;
                modemConfig_init[l] = key;
            }
        }

        if (modemConfig_cursor == 3)
        {
            l = strlen(modemConfig_hangup);
            if (l < 15)
            {
                modemConfig_hangup[l+1] = 0;
                modemConfig_hangup[l] = key;
            }
        }
    }
}

//=============================================================================
/* LAN CONFIG MENU */

int		lanConfig_cursor = -1;
int		lanConfig_cursor_table [] = {72, 92, 124};
#define NUM_LANCONFIG_CMDS	3

int 	lanConfig_port;
char	lanConfig_portname[6];
char	lanConfig_joinname[22];

void M_Menu_LanConfig_f (void)
{
    key_dest = key_menu;
    m_state = m_lanconfig;
    m_entersound = true;
    if (lanConfig_cursor == -1)
    {
        if (JoiningGame && TCPIPConfig)
            lanConfig_cursor = 2;
        else
            lanConfig_cursor = 1;
    }
    if (StartingGame && lanConfig_cursor == 2)
        lanConfig_cursor = 1;
    lanConfig_port = DEFAULTnet_hostport;
    sprintf(lanConfig_portname, "%u", lanConfig_port);

    m_return_onerror = false;
    m_return_reason[0] = 0;
}


void M_LanConfig_Draw (void)
{
    qpic_t	*p;
    int		basex;
    char	*startJoin;
    char	*protocol;

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    basex = (320-p->width)/2;
    M_DrawPic (basex, 4, p);

    if (StartingGame)
        startJoin = "New Game";
    else startJoin = "Join Game";

    if (IPXConfig)
        protocol = "IPX";
    else protocol = "TCP/IP";

    M_Print (basex, 32, va ("%s - %s", startJoin, protocol));
    basex -= 8;		// line them up correctly

    M_Print (basex + 8, 52, "Address:");

    if (IPXConfig)
        M_Print (basex+9*8, 52, my_ipx_address);
    else M_Print (basex+9*8, 52, my_tcpip_address);

    M_Print (basex + 8, lanConfig_cursor_table[0], "Port");
    M_DrawTextBox (basex+8*8 + 16, lanConfig_cursor_table[0]-8, 6, 1);
    M_Print (basex+9*8, lanConfig_cursor_table[0], lanConfig_portname);

    if (JoiningGame)
    {
        M_Print (basex, lanConfig_cursor_table[1], "Search for local games...");
        M_Print (basex, 108, "Join game at:");
        M_DrawTextBox (basex+24, lanConfig_cursor_table[2]-8, 22, 1);
        M_Print (basex+16, lanConfig_cursor_table[2], lanConfig_joinname);
    }
    else
    {
        M_DrawTextBox (basex + 16, lanConfig_cursor_table[1]-8, 2, 1);
        M_Print (basex+8, lanConfig_cursor_table[1], "OK");
    }

    M_DrawCharacter (basex - 4, lanConfig_cursor_table [lanConfig_cursor], 12+((int)(realtime*4)&1));

    if (lanConfig_cursor == 0)
        M_DrawCharacter (basex+9*8 + 8 + 8*strlen(lanConfig_portname), lanConfig_cursor_table [0], 10+((int)(realtime*4)&1));

    if (lanConfig_cursor == 2)
        M_DrawCharacter (basex+24 + 8*strlen(lanConfig_joinname), lanConfig_cursor_table [2], 10+((int)(realtime*4)&1));

    if (*m_return_reason)
        M_PrintWhite (basex, 148, m_return_reason);
}


void M_LanConfig_Key (int key)
{
    int		l;

    switch (key)
    {
    case K_ESCAPE:
        M_Menu_Net_f ();
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        lanConfig_cursor--;
        if (lanConfig_cursor < 0)
            lanConfig_cursor = NUM_LANCONFIG_CMDS-1;
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        lanConfig_cursor++;
        if (lanConfig_cursor >= NUM_LANCONFIG_CMDS)
            lanConfig_cursor = 0;
        break;

    case K_ENTER:
        if (lanConfig_cursor == 0)
            break;

        m_entersound = true;

        M_ConfigureNetSubsystem ();

        if (lanConfig_cursor == 1)
        {
            if (StartingGame)
            {
                M_Menu_GameOptions_f ();
                break;
            }

            M_Menu_Search_f();
            break;
        }

        if (lanConfig_cursor == 2)
        {
            m_return_state = m_state;
            m_return_onerror = true;
            key_dest = key_game;
            m_state = m_none;
            Cbuf_AddText ( va ("connect \"%s\"\n", lanConfig_joinname) );
            break;
        }

        break;

    case K_BACKSPACE:
        if (lanConfig_cursor == 0)
        {
            if (strlen(lanConfig_portname))
                lanConfig_portname[strlen(lanConfig_portname)-1] = 0;
        }

        if (lanConfig_cursor == 2)
        {
            if (strlen(lanConfig_joinname))
                lanConfig_joinname[strlen(lanConfig_joinname)-1] = 0;
        }
        break;

    default:
        if (key < 32 || key > 127)
            break;

        if (lanConfig_cursor == 2)
        {
            l = strlen(lanConfig_joinname);
            if (l < 21)
            {
                lanConfig_joinname[l+1] = 0;
                lanConfig_joinname[l] = key;
            }
        }

        if (key < '0' || key > '9')
            break;

        if (lanConfig_cursor == 0)
        {
            l = strlen(lanConfig_portname);
            if (l < 5)
            {
                lanConfig_portname[l+1] = 0;
                lanConfig_portname[l] = key;
            }
        }
    }

    if (StartingGame && lanConfig_cursor == 2)
        if (key == K_UPARROW)
            lanConfig_cursor = 1;
        else lanConfig_cursor = 0;

    l =  atoi(lanConfig_portname);

    if (l > 65535)
        l = lanConfig_port;
    else lanConfig_port = l;

    sprintf(lanConfig_portname, "%u", lanConfig_port);
}

//=============================================================================
/* GAME OPTIONS MENU */

typedef struct
{
    char	*name;
    char	*description;
} level_t;

level_t		levels[] =
{
    {"start", "Entrance"},	// 0

    {"e1m1", "Slipgate Complex"},				// 1
    {"e1m2", "Castle of the Damned"},
    {"e1m3", "The Necropolis"},
    {"e1m4", "The Grisly Grotto"},
    {"e1m5", "Gloom Keep"},
    {"e1m6", "The Door To Chthon"},
    {"e1m7", "The House of Chthon"},
    {"e1m8", "Ziggurat Vertigo"},

    {"e2m1", "The Installation"},				// 9
    {"e2m2", "Ogre Citadel"},
    {"e2m3", "Crypt of Decay"},
    {"e2m4", "The Ebon Fortress"},
    {"e2m5", "The Wizard's Manse"},
    {"e2m6", "The Dismal Oubliette"},
    {"e2m7", "Underearth"},

    {"e3m1", "Termination Central"},			// 16
    {"e3m2", "The Vaults of Zin"},
    {"e3m3", "The Tomb of Terror"},
    {"e3m4", "Satan's Dark Delight"},
    {"e3m5", "Wind Tunnels"},
    {"e3m6", "Chambers of Torment"},
    {"e3m7", "The Haunted Halls"},

    {"e4m1", "The Sewage System"},				// 23
    {"e4m2", "The Tower of Despair"},
    {"e4m3", "The Elder God Shrine"},
    {"e4m4", "The Palace of Hate"},
    {"e4m5", "Hell's Atrium"},
    {"e4m6", "The Pain Maze"},
    {"e4m7", "Azure Agony"},
    {"e4m8", "The Nameless City"},

    {"end", "Shub-Niggurath's Pit"},			// 31

    {"dm1", "Place of Two Deaths"},				// 32
    {"dm2", "Claustrophobopolis"},
    {"dm3", "The Abandoned Base"},
    {"dm4", "The Bad Place"},
    {"dm5", "The Cistern"},
    {"dm6", "The Dark Zone"}
};

//MED 01/06/97 added hipnotic levels
level_t     hipnoticlevels[] =
{
    {"start", "Command HQ"},  // 0

    {"hip1m1", "The Pumping Station"},          // 1
    {"hip1m2", "Storage Facility"},
    {"hip1m3", "The Lost Mine"},
    {"hip1m4", "Research Facility"},
    {"hip1m5", "Military Complex"},

    {"hip2m1", "Ancient Realms"},          // 6
    {"hip2m2", "The Black Cathedral"},
    {"hip2m3", "The Catacombs"},
    {"hip2m4", "The Crypt"},
    {"hip2m5", "Mortum's Keep"},
    {"hip2m6", "The Gremlin's Domain"},

    {"hip3m1", "Tur Torment"},       // 12
    {"hip3m2", "Pandemonium"},
    {"hip3m3", "Limbo"},
    {"hip3m4", "The Gauntlet"},

    {"hipend", "Armagon's Lair"},       // 16

    {"hipdm1", "The Edge of Oblivion"}           // 17
};

//PGM 01/07/97 added rogue levels
//PGM 03/02/97 added dmatch level
level_t		roguelevels[] =
{
    {"start",	"Split Decision"},
    {"r1m1",	"Deviant's Domain"},
    {"r1m2",	"Dread Portal"},
    {"r1m3",	"Judgement Call"},
    {"r1m4",	"Cave of Death"},
    {"r1m5",	"Towers of Wrath"},
    {"r1m6",	"Temple of Pain"},
    {"r1m7",	"Tomb of the Overlord"},
    {"r2m1",	"Tempus Fugit"},
    {"r2m2",	"Elemental Fury I"},
    {"r2m3",	"Elemental Fury II"},
    {"r2m4",	"Curse of Osiris"},
    {"r2m5",	"Wizard's Keep"},
    {"r2m6",	"Blood Sacrifice"},
    {"r2m7",	"Last Bastion"},
    {"r2m8",	"Source of Evil"},
    {"ctf1",    "Division of Change"}
};

typedef struct
{
    char	*description;
    int		firstLevel;
    int		levels;
} episode_t;

episode_t	episodes[] =
{
    {"Welcome to Quake", 0, 1},
    {"Doomed Dimension", 1, 8},
    {"Realm of Black Magic", 9, 7},
    {"Netherworld", 16, 7},
    {"The Elder World", 23, 8},
    {"Final Level", 31, 1},
    {"Deathmatch Arena", 32, 6}
};

//MED 01/06/97  added hipnotic episodes
episode_t   hipnoticepisodes[] =
{
    {"Scourge of Armagon", 0, 1},
    {"Fortress of the Dead", 1, 5},
    {"Dominion of Darkness", 6, 6},
    {"The Rift", 12, 4},
    {"Final Level", 16, 1},
    {"Deathmatch Arena", 17, 1}
};

//PGM 01/07/97 added rogue episodes
//PGM 03/02/97 added dmatch episode
episode_t	rogueepisodes[] =
{
    {"Introduction", 0, 1},
    {"Hell's Fortress", 1, 7},
    {"Corridors of Time", 8, 8},
    {"Deathmatch Arena", 16, 1}
};

int	startepisode;
int	startlevel;
int maxplayers;
qboolean m_serverInfoMessage = false;
double m_serverInfoMessageTime;

void M_Menu_GameOptions_f (void)
{
    key_dest = key_menu;
    m_state = m_gameoptions;
    m_entersound = true;
    if (maxplayers == 0)
        maxplayers = svs.maxclients;
    if (maxplayers < 2)
        maxplayers = svs.maxclientslimit;
}


int gameoptions_cursor_table[] = {40, 56, 64, 72, 80, 88, 96, 112, 120};
#define	NUM_GAMEOPTIONS	9
int		gameoptions_cursor;

void R_MapShot (char *mapname);
extern qboolean r_mapshot;
extern int mapshot_texturenum;
void Draw_Image (int image_texnum, float x0, float y0, float x1, float y1);

void M_GameOptions_Draw (void)
{
    qpic_t	*p;
    int		x;
    char mapname[64];

    M_DrawTransPic2 (8, 8, Draw_CachePic ("gfx/qplaque.lmp") );
    p = Draw_CachePic ("gfx/p_multi.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    M_DrawTextBox (152, 32, 10, 1);
    M_Print (144, 40, "begin game");

    M_Print (0, 56, "    Max players");
    M_Print (144, 56, va("%i", maxplayers) );

    M_Print (0, 64, "      Game Type");
    if (coop.value)
        M_Print (144, 64, "Cooperative");
    else
        M_Print (144, 64, "Deathmatch");

    M_Print (0, 72, "       Teamplay");

    if (rogue)
    {
        char *msg;

        switch((int)teamplay.value)
        {
        case 1:
            msg = "No Friendly Fire";
            break;
        case 2:
            msg = "Friendly Fire";
            break;
        case 3:
            msg = "Tag";
            break;
        case 4:
            msg = "Capture the Flag";
            break;
        case 5:
            msg = "One Flag CTF";
            break;
        case 6:
            msg = "Three Team CTF";
            break;
        default:
            msg = "Off";
            break;
        }

        M_Print (144, 72, msg);
    }
    else
    {
        char *msg;

        switch((int)teamplay.value)
        {
        case 1:
            msg = "No Friendly Fire";
            break;
        case 2:
            msg = "Friendly Fire";
            break;
        default:
            msg = "Off";
            break;
        }

        M_Print (144, 72, msg);
    }

    M_Print (0, 80, "          Skill");

    if (skill.value == 0)
        M_Print (144, 80, "Easy");
    else if (skill.value == 1)
        M_Print (144, 80, "Normal");
    else if (skill.value == 2)
        M_Print (144, 80, "Hard");
    else M_Print (144, 80, "Nightmare");

    M_Print (0, 88, "     Frag Limit");
    if (fraglimit.value == 0)
        M_Print (144, 88, "none");
    else
        M_Print (144, 88, va("%i frags", (int)fraglimit.value));

    M_Print (0, 96, "     Time Limit");
    if (timelimit.value == 0)
        M_Print (144, 96, "none");
    else
        M_Print (144, 96, va("%i minutes", (int)timelimit.value));

    M_Print (0, 112, "        Episode");

    if (hipnotic)
        M_Print (144, 112, hipnoticepisodes[startepisode].description);
    else if (rogue)
        M_Print (144, 112, rogueepisodes[startepisode].description);
    else
        M_Print (144, 112, episodes[startepisode].description);

    M_Print (0, 120, "          Level");

    if (hipnotic)
    {
        M_Print (144, 120, hipnoticlevels[hipnoticepisodes[startepisode].firstLevel + startlevel].description);
        strcpy (mapname, hipnoticlevels[hipnoticepisodes[startepisode].firstLevel + startlevel].name);
    }
    else if (rogue)
    {
        M_Print (144, 120, roguelevels[rogueepisodes[startepisode].firstLevel + startlevel].description);
        strcpy (mapname, roguelevels[rogueepisodes[startepisode].firstLevel + startlevel].name);
    }
    else
    {
        M_Print (144, 120, levels[episodes[startepisode].firstLevel + startlevel].description);
        strcpy (mapname, levels[episodes[startepisode].firstLevel + startlevel].name);
    }

    // line cursor
    M_DrawCharacter (136, gameoptions_cursor_table[gameoptions_cursor], 12+((int)(realtime*4)&1));

    // draw mapshot for current level here
    R_MapShot (mapname);

    if (r_mapshot)
    {
        int x0, x1;

        x0 = (vid.width - 192) / 2;
        x1 = x0 + 192;

        Draw_Image (mapshot_texturenum, x0, 132, x1, 276);
        r_mapshot = false;
    }
    else M_PrintWhite (32, 194, "No Mapshot Available");

    // do this last cos we always want to see it
    if (m_serverInfoMessage)
    {
        if ((realtime - m_serverInfoMessageTime) < 5.0)
        {
            x = (320-26*8)/2;

            // center the text
            M_DrawTextBox (x + 12, 138, 24, 4);
            x += 8;
            M_Print (x, 146, "  More than 4 players   ");
            M_Print (x, 154, " requires using command ");
            M_Print (x, 162, "line parameters; please ");
            M_Print (x, 170, "   see techinfo.txt.    ");
        }
        else m_serverInfoMessage = false;
    }
}


void M_NetStart_Change (int dir)
{
    int count;

    switch (gameoptions_cursor)
    {
    case 1:
        maxplayers += dir;
        if (maxplayers > svs.maxclientslimit)
        {
            maxplayers = svs.maxclientslimit;
            m_serverInfoMessage = true;
            m_serverInfoMessageTime = realtime;
        }
        if (maxplayers < 2)
            maxplayers = 2;
        break;

    case 2:
        Cvar_SetValue ("coop", coop.value ? 0 : 1);
        break;

    case 3:
        if (rogue)
            count = 6;
        else
            count = 2;

        Cvar_SetValue ("teamplay", teamplay.value + dir);
        if (teamplay.value > count)
            Cvar_SetValue ("teamplay", 0);
        else if (teamplay.value < 0)
            Cvar_SetValue ("teamplay", count);
        break;

    case 4:
        Cvar_SetValue ("skill", skill.value + dir);
        if (skill.value > 3)
            Cvar_SetValue ("skill", 0);
        if (skill.value < 0)
            Cvar_SetValue ("skill", 3);
        break;

    case 5:
        Cvar_SetValue ("fraglimit", fraglimit.value + dir*10);
        if (fraglimit.value > 100)
            Cvar_SetValue ("fraglimit", 0);
        if (fraglimit.value < 0)
            Cvar_SetValue ("fraglimit", 100);
        break;

    case 6:
        Cvar_SetValue ("timelimit", timelimit.value + dir*5);
        if (timelimit.value > 60)
            Cvar_SetValue ("timelimit", 0);
        if (timelimit.value < 0)
            Cvar_SetValue ("timelimit", 60);
        break;

    case 7:
        startepisode += dir;
        //MED 01/06/97 added hipnotic count
        if (hipnotic)
            count = 6;
        //PGM 01/07/97 added rogue count
        //PGM 03/02/97 added 1 for dmatch episode
        else if (rogue)
            count = 4;
        else if (registered.value)
            count = 7;
        else
            count = 2;

        if (startepisode < 0)
            startepisode = count - 1;

        if (startepisode >= count)
            startepisode = 0;

        startlevel = 0;
        break;

    case 8:
        startlevel += dir;
        //MED 01/06/97 added hipnotic episodes
        if (hipnotic)
            count = hipnoticepisodes[startepisode].levels;
        //PGM 01/06/97 added hipnotic episodes
        else if (rogue)
            count = rogueepisodes[startepisode].levels;
        else
            count = episodes[startepisode].levels;

        if (startlevel < 0)
            startlevel = count - 1;

        if (startlevel >= count)
            startlevel = 0;
        break;
    }
}

void M_GameOptions_Key (int key)
{
    switch (key)
    {
    case K_ESCAPE:
        M_Menu_Net_f ();
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        gameoptions_cursor--;
        if (gameoptions_cursor < 0)
            gameoptions_cursor = NUM_GAMEOPTIONS-1;
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        gameoptions_cursor++;
        if (gameoptions_cursor >= NUM_GAMEOPTIONS)
            gameoptions_cursor = 0;
        break;

    case K_LEFTARROW:
        if (gameoptions_cursor == 0)
            break;
        S_LocalSound ("misc/menu3.wav");
        M_NetStart_Change (-1);
        break;

    case K_RIGHTARROW:
        if (gameoptions_cursor == 0)
            break;
        S_LocalSound ("misc/menu3.wav");
        M_NetStart_Change (1);
        break;

    case K_ENTER:
        S_LocalSound ("misc/menu2.wav");
        if (gameoptions_cursor == 0)
        {
            if (sv.active)
                Cbuf_AddText ("disconnect\n");
            Cbuf_AddText ("listen 0\n");	// so host_netport will be re-examined
            Cbuf_AddText ( va ("maxplayers %u\n", maxplayers) );
            SCR_BeginLoadingPlaque ();

            if (hipnotic)
                Cbuf_AddText ( va ("map %s\n", hipnoticlevels[hipnoticepisodes[startepisode].firstLevel + startlevel].name) );
            else if (rogue)
                Cbuf_AddText ( va ("map %s\n", roguelevels[rogueepisodes[startepisode].firstLevel + startlevel].name) );
            else
                Cbuf_AddText ( va ("map %s\n", levels[episodes[startepisode].firstLevel + startlevel].name) );

            return;
        }

        M_NetStart_Change (1);
        break;
    }
}

//=============================================================================
/* SEARCH MENU */

qboolean	searchComplete = false;
double		searchCompleteTime;

void M_Menu_Search_f (void)
{
    key_dest = key_menu;
    m_state = m_search;
    m_entersound = false;
    slistSilent = true;
    slistLocal = false;
    searchComplete = false;
    NET_Slist_f();

}


void M_Search_Draw (void)
{
    qpic_t	*p;
    int x;

    p = Draw_CachePic ("gfx/p_multi.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);
    x = (320/2) - ((12*8)/2) + 4;
    M_DrawTextBox (x-8, 32, 12, 1);
    M_Print (x - 16, 40, "Searching...");

    if(slistInProgress)
    {
        NET_Poll();
        return;
    }

    if (! searchComplete)
    {
        searchComplete = true;
        searchCompleteTime = realtime;
    }

    if (hostCacheCount)
    {
        M_Menu_ServerList_f ();
        return;
    }

    M_PrintWhite ((320/2) - ((22*8)/2), 64, "No Quake servers found");
    if ((realtime - searchCompleteTime) < 3.0)
        return;

    M_Menu_LanConfig_f ();
}


void M_Search_Key (int key)
{
}

//=============================================================================
/* SLIST MENU */

int		slist_cursor;
qboolean slist_sorted;

void M_Menu_ServerList_f (void)
{
    key_dest = key_menu;
    m_state = m_slist;
    m_entersound = true;
    slist_cursor = 0;
    m_return_onerror = false;
    m_return_reason[0] = 0;
    slist_sorted = false;
}


void M_ServerList_Draw (void)
{
    int		n;
    char	string [64];
    qpic_t	*p;

    if (!slist_sorted)
    {
        if (hostCacheCount > 1)
        {
            int	i,j;
            hostcache_t temp;
            for (i = 0; i < hostCacheCount; i++)
                for (j = i+1; j < hostCacheCount; j++)
                    if (strcmp(hostcache[j].name, hostcache[i].name) < 0)
                    {
                        memcpy(&temp, &hostcache[j], sizeof(hostcache_t));
                        memcpy(&hostcache[j], &hostcache[i], sizeof(hostcache_t));
                        memcpy(&hostcache[i], &temp, sizeof(hostcache_t));
                    }
        }
        slist_sorted = true;
    }

    p = Draw_CachePic ("gfx/p_multi.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);
    for (n = 0; n < hostCacheCount; n++)
    {
        if (hostcache[n].maxusers)
            sprintf(string, "%-15.15s %-15.15s %2u/%2u\n", hostcache[n].name, hostcache[n].map, hostcache[n].users, hostcache[n].maxusers);
        else
            sprintf(string, "%-15.15s %-15.15s\n", hostcache[n].name, hostcache[n].map);
        M_Print (16, 32 + 8*n, string);
    }
    M_DrawCharacter (8, 32 + slist_cursor*8, 12+((int)(realtime*4)&1));

    if (*m_return_reason)
        M_PrintWhite (16, 148, m_return_reason);
}


void M_ServerList_Key (int k)
{
    switch (k)
    {
    case K_ESCAPE:
        M_Menu_LanConfig_f ();
        break;

    case K_SPACE:
        M_Menu_Search_f ();
        break;

    case K_UPARROW:
    case K_LEFTARROW:
        S_LocalSound ("misc/menu1.wav");
        slist_cursor--;
        if (slist_cursor < 0)
            slist_cursor = hostCacheCount - 1;
        break;

    case K_DOWNARROW:
    case K_RIGHTARROW:
        S_LocalSound ("misc/menu1.wav");
        slist_cursor++;
        if (slist_cursor >= hostCacheCount)
            slist_cursor = 0;
        break;

    case K_ENTER:
        S_LocalSound ("misc/menu2.wav");
        m_return_state = m_state;
        m_return_onerror = true;
        slist_sorted = false;
        key_dest = key_game;
        m_state = m_none;
        Cbuf_AddText ( va ("connect \"%s\"\n", hostcache[slist_cursor].cname) );
        break;

    default:
        break;
    }

}

//=============================================================================
/* Menu Subsystem */


void M_Init (void)
{
    Cmd_AddCommand ("togglemenu", M_ToggleMenu_f);

    Cmd_AddCommand ("menu_main", M_Menu_Main_f);
    Cmd_AddCommand ("menu_singleplayer", M_Menu_SinglePlayer_f);
    Cmd_AddCommand ("menu_load", M_Menu_Load_f);
    Cmd_AddCommand ("menu_save", M_Menu_Save_f);
    Cmd_AddCommand ("menu_multiplayer", M_Menu_MultiPlayer_f);
    Cmd_AddCommand ("menu_setup", M_Menu_Setup_f);
    Cmd_AddCommand ("menu_options", M_Menu_Options_f);
    Cmd_AddCommand ("menu_keys", M_Menu_Keys_f);
    Cmd_AddCommand ("help", M_Menu_Help_f);
    Cmd_AddCommand ("menu_quit", M_Menu_Quit_f);
}


void M_Draw (void)
{
    if (m_state == m_none || key_dest != key_menu)
        return;

    if (!m_recursiveDraw)
    {
        scr_copyeverything = 1;

        if (scr_con_current)
        {
            Draw_ConsoleBackground (vid.height);
            VID_UnlockBuffer ();
            S_ExtraUpdate ();
            VID_LockBuffer ();
        }
        else M_DrawTextBox (0, 0, 40, 34); // Draw_FadeScreen ();

        scr_fullupdate = 0;
    }
    else
    {
        m_recursiveDraw = false;
    }

    switch (m_state)
    {
    case m_none:
        break;

    case m_main:
        M_Main_Draw ();
        break;

    case m_singleplayer:
        M_SinglePlayer_Draw ();
        break;

    case m_load:
        M_Load_Draw ();
        break;

    case m_save:
        M_Save_Draw ();
        break;

    case m_multiplayer:
        M_MultiPlayer_Draw ();
        break;

    case m_setup:
        M_Setup_Draw ();
        break;

    case m_net:
        M_Net_Draw ();
        break;

    case m_options:
        M_Options_Draw (1, &options_store[0]);
        break;

    case m_options_general:
        M_General_Draw (&general_store[0]);
        break;

    case m_options_basic:
        M_Basic_Draw (&basic_store[0]);
        break;

    case m_options_static:
        M_Static_Draw (&static_store[0]);
        break;

    case m_options_dynamic:
        M_Dynamic_Draw (&dynamic_store[0]);
        break;

    case m_options_candy:
        M_Candy_Draw (&candy_store[0]);
        break;

    case m_options_hardcore:
        M_HardCore_Draw (&hardcore_store[0]);
        break;

    case m_options_fog:
        M_Fog_Draw (&fog_store[0]);
        break;

    case m_keys:
        M_Keys_Draw ();
        break;

    case m_help:
        M_Help_Draw ();
        break;

    case m_quit:
        M_Quit_Draw ();
        break;

    case m_serialconfig:
        M_SerialConfig_Draw ();
        break;

    case m_modemconfig:
        M_ModemConfig_Draw ();
        break;

    case m_lanconfig:
        M_LanConfig_Draw ();
        break;

    case m_gameoptions:
        M_GameOptions_Draw ();
        break;

    case m_search:
        M_Search_Draw ();
        break;

    case m_slist:
        M_ServerList_Draw ();
        break;
    }

    if (m_entersound)
    {
        S_LocalSound ("misc/menu2.wav");
        m_entersound = false;
    }

    VID_UnlockBuffer ();
    S_ExtraUpdate ();
    VID_LockBuffer ();
}


void M_Keydown (int key)
{
    switch (m_state)
    {
    case m_none:
        return;

    case m_main:
        M_Main_Key (key);
        return;

    case m_singleplayer:
        M_SinglePlayer_Key (key);
        return;

    case m_load:
        M_Load_Key (key);
        return;

    case m_save:
        M_Save_Key (key);
        return;

    case m_multiplayer:
        M_MultiPlayer_Key (key);
        return;

    case m_setup:
        M_Setup_Key (key);
        return;

    case m_net:
        M_Net_Key (key);
        return;

    case m_options:
        options_cursor = M_SubMenu_Key (key, options_cursor, &options_store[0]);
        return;

    case m_options_general:
        g_options_cursor = M_SubMenu_Key (key, g_options_cursor, &general_store[0]);
        break;

    case m_options_basic:
        b_options_cursor = M_SubMenu_Key (key, b_options_cursor, &basic_store[0]);
        break;

    case m_options_static:
        s_options_cursor = M_SubMenu_Key (key, s_options_cursor, &static_store[0]);
        break;

    case m_options_dynamic:
        d_options_cursor = M_SubMenu_Key (key, d_options_cursor, &dynamic_store[0]);
        break;

    case m_options_candy:
        c_options_cursor = M_SubMenu_Key (key, c_options_cursor, &candy_store[0]);
        break;

    case m_options_hardcore:
        h_options_cursor = M_SubMenu_Key (key, h_options_cursor, &hardcore_store[0]);
        break;

    case m_options_fog:
        f_options_cursor = M_SubMenu_Key (key, f_options_cursor, &fog_store[0]);
        break;

    case m_keys:
        M_Keys_Key (key);
        return;

    case m_help:
        M_Help_Key (key);
        return;

    case m_quit:
        M_Quit_Key (key);
        return;

    case m_serialconfig:
        M_SerialConfig_Key (key);
        return;

    case m_modemconfig:
        M_ModemConfig_Key (key);
        return;

    case m_lanconfig:
        M_LanConfig_Key (key);
        return;

    case m_gameoptions:
        M_GameOptions_Key (key);
        return;

    case m_search:
        M_Search_Key (key);
        break;

    case m_slist:
        M_ServerList_Key (key);
        return;
    }
}


void M_ConfigureNetSubsystem(void)
{
    // enable/disable net systems to match desired config
    Cbuf_AddText ("stopdemo\n");

    if (SerialConfig || DirectConfig)
        Cbuf_AddText ("com1 enable\n");

    if (IPXConfig || TCPIPConfig)
        net_hostport = lanConfig_port;
}


// place at the bottom cos we need all the possible functions it could call declared above it
void M_MenuItem_Draw (float menu_cvar, mitem_type_t mitem_type, char *mitem_string,
                      char *m_switch_1, char *m_switch_2, char *m_switch_3, cvar_t *menu_store,
                      float scale_min, float scale_max, menu_store_t *store, reload_t reload,
                      float step, void (*submenu_func) ())
{
    // bit of error checking...
    if (submenu_items > MAX_MENU_STORE) Sys_Error ("Ask MH to increase the menu stores!\n(or increase them yourself...)");

    // print the item name
    M_Print (192 - (strlen (mitem_string) * 8), current_submenu_cursor, mitem_string);

    // draw the value in one of 4 different ways depending on the type
    switch (mitem_type)
    {
    case on_off:
    case on_off2:
        M_DrawCheckbox (220, current_submenu_cursor, (int) menu_cvar, m_switch_1, m_switch_2);
        break;

    case three_item:
        M_DrawTripleCheckbox (220, current_submenu_cursor, (int) menu_cvar, m_switch_1, m_switch_2, m_switch_3);
        break;

    case slider:
        M_DrawSlider (220, current_submenu_cursor, menu_cvar);
        break;

    case scale:
        M_DrawNumericScale (220, current_submenu_cursor, menu_cvar);
        break;

    case sub_menu:
        // veeery powerful this - use with extreme caution...
        // draw nothing else here
        break;

    case spacer:
        M_Print (220, current_submenu_cursor, m_switch_1);
        break;

    default:
        break;
    }

    // store out the stuff
    store[submenu_items].cvar_store = menu_store;
    store[submenu_items].min_scale = scale_min;
    store[submenu_items].max_scale = scale_max;
    store[submenu_items].item_type = mitem_type;
    store[submenu_items].requires_reload = reload;
    store[submenu_items].slider_step = step;
    store[submenu_items].submenu_func = submenu_func;

    // finish up
    submenu_items++;
    current_submenu_cursor += 8;
}
