/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// texture management routines

#include "quakedef.h"
#include "gl_textures.h"

#ifdef USE_JPEG
#include "jpeglib.h"
#endif


extern int QPalR[256];
extern int QPalG[256];
extern int QPalB[256];

extern int detailmip;
extern cvar_t gl_picmip;
extern int oldtexture;

extern int gl_filter_min;
extern int gl_filter_max;

extern cvar_t gl_detailtextures;
extern cvar_t gl_cachereplace;

/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
    int		i;
    gl_texture_t	*glt;

    for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
    {
        if (!strcmp (identifier, glt->identifier))
            return gltextures[i].texnum;
    }

    return -1;
}

// returns the full structure rather than just a texture num
gl_texture_t *GL_FindFullTexture (char *identifier)
{
    int		i;
    gl_texture_t	*glt;

    for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
    {
        if (!strcmp (identifier, glt->identifier))
            return &gltextures[i];
    }

    return NULL;
}

/*
================
GL_ResampleTexture
================
*/
void GL_ResampleTexture (unsigned *in, int inwidth, int inheight, unsigned *out,  int outwidth, int outheight)
{
    int		i, j;
    unsigned	*inrow;
    unsigned	frac, fracstep;

    fracstep = inwidth * 0x10000 / outwidth;

    for (i = 0; i < outheight; i++, out += outwidth)
    {
        inrow = in + inwidth * (i * inheight / outheight);
        frac = fracstep >> 1;

        for (j = 0; j < outwidth; j += 4)
        {
            out[j] = inrow[frac>>16];
            frac += fracstep;

            out[j + 1] = inrow[frac>>16];
            frac += fracstep;

            out[j + 2] = inrow[frac>>16];
            frac += fracstep;

            out[j + 3] = inrow[frac>>16];
            frac += fracstep;
        }
    }
}


/*
================
GL_MipMap

Operates in place, quartering the size of the texture
================
*/

// i'm probably missing a builtin to do this, but it's more fun to code it myself
byte qround (float num)
{
    double frac_part, int_part;

    frac_part = modf ((double) num, &int_part);

    if (frac_part >= 0.5)
    {
        int_part = int_part + 1;
    }
    return (byte) int_part;
}


void GL_MipMap (byte *in, int width, int height)
{
    int	i, j;
    byte *out;

    width <<= 2;
    height >>= 1;
    out = in;

    if (detailmip)
    {
        for (i = 0; i < height; i++, in += width)
        {
            for (j = 0; j < width; j += 8, out += 4, in += 8)
            {
                // float operations - somewhat slower but more accurate
                out[0] = qround ((float)(in[0] + in[4] + in[width + 0] + in[width + 4]) / 4.0);
                out[1] = qround ((float)(in[1] + in[5] + in[width + 1] + in[width + 5]) / 4.0);
                out[2] = qround ((float)(in[2] + in[6] + in[width + 2] + in[width + 6]) / 4.0);
                out[3] = qround ((float)(in[3] + in[7] + in[width + 3] + in[width + 7]) / 4.0);
            }
        }

        return;
    }

    for (i = 0; i < height; i++, in += width)
    {
        for (j = 0; j < width; j += 8, out += 4, in += 8)
        {
            // faster but less accurate
            out[0] = (in[0] + in[4] + in[width + 0] + in[width + 4]) >> 2;
            out[1] = (in[1] + in[5] + in[width + 1] + in[width + 5]) >> 2;
            out[2] = (in[2] + in[6] + in[width + 2] + in[width + 6]) >> 2;
            out[3] = (in[3] + in[7] + in[width + 3] + in[width + 7]) >> 2;
        }
    }
}


/*
===============
GL_Upload32
===============
*/
extern int fog_colour[3];
extern int fog_density;

qboolean is_water;
extern int notexture_num;
extern cvar_t gl_round_down;

union rgba_u
{
    unsigned int rgba_i;
    unsigned char rgba_c[4];
};


int GL_Upload32 (unsigned *data, int width, int height, qboolean mipmap, qboolean alpha)
{
    int	samples;
    static unsigned	scaled[1024*1024];
    int scaled_width, scaled_height;
    int i;
    long int total_rgb[3];
    union rgba_u rgba;
    int max_fog;
    int	miplevel;
    static unsigned mip0[1024*1024];

    if (!hlbsp && is_water)
    {
        // halflife liquids already have their underwater fog colour and density set - for quake,
        // we must do a nasty hack...
        fog_density = 128;

        total_rgb[0] = total_rgb[1] = total_rgb[2] = 0;

        for (i = 0; i < (width * height); i++)
        {
            // using a union here is faster than a (byte *) cast
            rgba.rgba_i = data[i];

            total_rgb[0] += (long) rgba.rgba_c[0];
            total_rgb[1] += (long) rgba.rgba_c[1];
            total_rgb[2] += (long) rgba.rgba_c[2];
        }

        // because quake uses low rgb values in it's palette, we'll boost the results a little
        // probably missing an obvious way of doing this...
        max_fog = 0;

        for (i = 0; i < 3; i++) if (total_rgb[i] >= max_fog) max_fog = total_rgb[i];

        // this should never happen
        if (!max_fog) max_fog = 1;

        fog_colour[0] = (int) (255.0 * ((float) total_rgb[0] / (float) max_fog));
        fog_colour[1] = (int) (255.0 * ((float) total_rgb[1] / (float) max_fog));
        fog_colour[2] = (int) (255.0 * ((float) total_rgb[2] / (float) max_fog));

        // this should never happen but we must check anyway, in case some nutter makes an
        // all black liquid texture (it takes all sorts...)
        if (fog_colour[0] == 0 && fog_colour[1] == 0 && fog_colour[2] == 0)
        {
            fog_colour[0] = 1;
            fog_colour[1] = 1;
            fog_colour[2] = 1;
        }

        // if the sum is greater than 300, the fog colour will look too strong, so decrease
        // them... we're not worried if the total is less (it's probably deliberate) - this can
        // weaken certain fogs that are predominately one colour (slime, lava...) so we'll have
        // to check for that too. (if they're all greater than 48 it's a safe assumption)
        max_fog = fog_colour[0] + fog_colour[1] + fog_colour[2];

        if (max_fog > 300 && fog_colour[0] > 48 && fog_colour[1] > 48 && fog_colour[2] > 48)
        {
            fog_colour[0] = (int) ((float) (fog_colour[0] * 255) / (float) max_fog);
            fog_colour[1] = (int) ((float) (fog_colour[1] * 255) / (float) max_fog);
            fog_colour[2] = (int) ((float) (fog_colour[2] * 255) / (float) max_fog);
        }
    }

    // reduce texture size down to a power of 2 size
    if (mipmap)
    {
        // restored gl_rounddown, and made this code considerably cleaner
        if (gl_round_down.value)
        {
            if (width >= 1024) scaled_width = 1024;
            else if (width >= 512) scaled_width = 512;
            else if (width >= 256) scaled_width = 256;
            else if (width >= 128) scaled_width = 128;
            else if (width >= 64) scaled_width = 64;
            else scaled_width = 32;

            if (height >= 1024) scaled_height = 1024;
            else if (height >= 512) scaled_height = 512;
            else if (height >= 256) scaled_height = 256;
            else if (height >= 128) scaled_height = 128;
            else if (height >= 64) scaled_height = 64;
            else scaled_height = 32;
        }
        else
        {
            if (width > 512) scaled_width = 1024;
            else if (width > 256) scaled_width = 512;
            else if (width > 128) scaled_width = 256;
            else if (width > 64) scaled_width = 128;
            else if (width > 32) scaled_width = 64;
            else scaled_width = 32;

            if (height > 512) scaled_height = 1024;
            else if (height > 256) scaled_height = 512;
            else if (height > 128) scaled_height = 256;
            else if (height > 64) scaled_height = 128;
            else if (height > 32) scaled_height = 64;
            else scaled_height = 32;
        }

        // picmipping
        scaled_width >>= (int) gl_picmip.value;
        scaled_height >>= (int) gl_picmip.value;

        // lower limit
        if (scaled_width < 32) scaled_width = 32;
        if (scaled_height < 32) scaled_height = 32;
    }
    else
    {
        // make 2D objects more detailed
        if (width >= 256) scaled_width = 1024;
        else if (width >= 128) scaled_width = 512;
        else if (width >= 64) scaled_width = 256;
        else if (width >= 32) scaled_width = 128;
        else if (width >= 16) scaled_width = 64;
        else scaled_width = 32;

        if (height >= 256) scaled_height = 1024;
        else if (height >= 128) scaled_height = 512;
        else if (height >= 64) scaled_height = 256;
        else if (height >= 32) scaled_height = 128;
        else if (height >= 16) scaled_height = 64;
        else scaled_height = 32;
    }

    // clamp at max texture size for the driver (normally 1024, 256 for 3dfx)
    if (scaled_width > gl_max_size.value) scaled_width = gl_max_size.value;
    if (scaled_height > gl_max_size.value) scaled_height = gl_max_size.value;

    // because of the max sizes i have on scaled_width and scaled_height, and the new size i
    // have set scaled at, the error check that was here is no longer relevant

    samples = alpha ? gl_alpha_format : gl_solid_format;

    if (scaled_width == width && scaled_height == height)
    {
        if (!mipmap)
        {
            glTexImage2D (GL_TEXTURE_2D, 0, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
            goto done;
        }

        memcpy (scaled, data, width * height * 4);
    }
    else GL_ResampleTexture (data, width, height, scaled, scaled_width, scaled_height);

    glTexImage2D (GL_TEXTURE_2D, 0, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);

    if (mipmap)
    {
        miplevel = 0;

        while (scaled_width > 1 || scaled_height > 1)
        {
            GL_MipMap ((byte *) scaled, scaled_width, scaled_height);

            scaled_width >>= 1;
            scaled_height >>= 1;

            if (scaled_width < 1) scaled_width = 1;
            if (scaled_height < 1) scaled_height = 1;

            miplevel++;

            glTexImage2D (GL_TEXTURE_2D, miplevel, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
        }
    }

done:
    if (mipmap)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
    else glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_max);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);

    return 1;
}


/*
===============
GL_Upload8

MH - rewrote this function cos it was confusing
===============
*/
int GL_Upload8 (byte *data, int width, int height,  qboolean mipmap, qboolean alpha)
{
    int			i, j, s;
    byte *q_rgba;

    s = width * height;
    q_rgba = (byte *) malloc (width * height * 4);

    // if there are no transparent pixels, make it a 3 component
    // texture even if it was specified as otherwise
    alpha = false;

    // i've received reports of false positives on this - this should only be tested if there are
    // no transparent components in the texture - bit slower to upload, but more reliable.
    for (i = 0; i < s; i++)
        if (data[i] == 255) alpha = true;

    // made this safe - was a sys_error
    if ((s & 3) && !alpha)
        return 0;
    else
    {
        for (i = 0, j = 0; i < s; i++)
        {
            if (data[i] == 255 && alpha)
            {
                // set the correct transparency
                q_rgba[j++] = 0;
                q_rgba[j++] = 0;
                q_rgba[j++] = 0;
                q_rgba[j++] = 0;
            }
            else
            {
                q_rgba[j++] = QPalR[data[i]];
                q_rgba[j++] = QPalG[data[i]];
                q_rgba[j++] = QPalB[data[i]];
                q_rgba[j++] = 255;
            }
        }
    }

    i = GL_Upload32 ((unsigned int *)q_rgba, width, height, mipmap, alpha);

    free (q_rgba);

    return i;
}

/*
================
GL_LoadTexture
================
*/
int lhcsumtable[256];

int MHGL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int bitsperpixel)
{
    int			i, s, lhcsum;
    gl_texture_t	*glt;

    // mh - if we have exceeded the maximum number of textures, load up the default
    // r_notexturemip instead... please note that id's r_notexturemip doesn't work
    // in glquake, so i made made my own.
    if (numgltextures >= MAX_GLTEXTURES - 1)
        return notexture_num;

    // if bpp is invalid, use the notexture texture
    if (bitsperpixel == 8 || bitsperpixel == 32)
    {
        if (mipmap)
        {
            gl_filter_min = GL_LINEAR_MIPMAP_LINEAR;
            gl_filter_max = GL_LINEAR;
        }
        else
        {
            gl_filter_min = GL_LINEAR;
            gl_filter_max = GL_LINEAR;
        }

        // LordHavoc: do a checksum to confirm the data really is the same as previous
        // occurances. well this isn't exactly a checksum, it's better than that but
        // not following any standards.
        lhcsum = 0;
        s = width * height * (bitsperpixel / 8);
        for (i = 0; i < 256; i++) lhcsumtable[i] = i + 1;
        for (i = 0; i < s; i++) lhcsum += (lhcsumtable[data[i] & 255]++);

        // see if the texture is allready present
        if (identifier[0])
        {
            for (i=0, glt=gltextures ; i < numgltextures ; i++, glt++)
            {
                if (!strcmp (identifier, glt->identifier))
                {
                    // gl_cachereplace 1 always replaces textures in the cache - although in most
                    // most cases lh's 'checksum' should be more than sufficient.
                    if (gl_cachereplace.value) goto GL_LoadTexture_setup;

                    // replace cached water textures so underwater fogging can be set properly
                    if (is_water) goto GL_LoadTexture_setup;

                    // cache mismatch fix from lh
                    if (lhcsum != glt->lhcsum || width != glt->width || height != glt->height)
                    {
                        Con_DPrintf("GL_LoadTexture: cache mismatch, replacing old texture\n");

                        // drop out with glt pointing to the texture to replace
                        goto GL_LoadTexture_setup;
                    }

                    return glt->texnum;
                }
            }
        }

        // LordHavoc: this was an else condition, causing disasterous results,
        // whoever at id or threewave must've been half asleep...
        glt = &gltextures[numgltextures];
        numgltextures++;

        strcpy (glt->identifier, identifier);
        glt->texnum = texture_extension_number;
        glt->origtexnum = glt->texnum;
        texture_extension_number++;

        // LordHavoc: label to drop out of the loop into the setup code
GL_LoadTexture_setup:
        glt->lhcsum = lhcsum; // LordHavoc: used to verify textures are identical
        glt->width = width;
        glt->height = height;
        glt->mipmap = mipmap;
        glt->alpha = alpha;

        if (!isDedicated)
        {
            glBindTexture (GL_TEXTURE_2D, glt->texnum);
            oldtexture = glt->texnum;

            // always go looking for alpha with 8 bit textures - if there is none, quake
            // will set it up as a non-alpha texture anyway...
            if (bitsperpixel == 8)
                i = GL_Upload8 (data, width, height, mipmap, true);
            else i = GL_Upload32 ((unsigned int *)data, width, height, mipmap, alpha);

            glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        }
        else i = 1;	// always assume success cos dedicated servers don't upload

        // if the upload failed, use the notexture texture
        if (i) return glt->texnum;
    }

    return notexture_num;
}


#ifdef USE_EXTTEX
/*
=======================
PCX header
=======================
*/
typedef struct pcx_s
{
    char	manufacturer;
    char	version;
    char	encoding;
    char	bits_per_pixel;
    unsigned short	xmin,ymin,xmax,ymax;
    unsigned short	hres,vres;
    unsigned char	palette[48];
    char	reserved;
    char	color_planes;
    unsigned short	bytes_per_line;
    unsigned short	palette_type;
    char	filler[58];
    unsigned 	data;			// unbounded
} pcx_t;

/*
=======================
TGA header
=======================
*/
#define TGA_MAXCOLORS 16384

/* Definitions for image types. */
#define TGA_Null	0	/* no image data */
#define TGA_Map		1	/* Uncompressed, color-mapped images. */
#define TGA_RGB		2	/* Uncompressed, RGB images. */
#define TGA_Mono	3	/* Uncompressed, black and white images. */
#define TGA_RLEMap	9	/* Runlength encoded color-mapped images. */
#define TGA_RLERGB	10	/* Runlength encoded RGB images. */
#define TGA_RLEMono	11	/* Compressed, black and white images. */
#define TGA_CompMap	32	/* Compressed color-mapped data, using Huffman, Delta, and runlength encoding. */
#define TGA_CompMap4	33	/* Compressed color-mapped data, using Huffman, Delta, and runlength encoding. 4-pass quadtree-type process. */

/* Definitions for interleave flag. */
#define TGA_IL_None	0	/* non-interleaved. */
#define TGA_IL_Two	1	/* two-way (even/odd) interleaving */
#define TGA_IL_Four	2	/* four way interleaving */
#define TGA_IL_Reserved	3	/* reserved */

/* Definitions for origin flag */
#define TGA_O_UPPER	0	/* Origin in lower left-hand corner. */
#define TGA_O_LOWER	1	/* Origin in upper left-hand corner. */

typedef struct tga_s
{
    unsigned char id_length;
    unsigned char colormap_type;
    unsigned char image_type;
    unsigned short colormap_index;
    unsigned short colormap_length;
    unsigned char colormap_size;
    unsigned short x_origin;
    unsigned short y_origin;
    unsigned short width;
    unsigned short height;
    unsigned char pixel_size;
    unsigned char attributes;
} tga_t;

int fgetLittleShort (FILE *f)
{
    byte	b1, b2;

    b1 = fgetc(f);
    b2 = fgetc(f);

    return (short)(b1 + b2*256);
}

int fgetLittleLong (FILE *f)
{
    byte	b1, b2, b3, b4;

    b1 = fgetc(f);
    b2 = fgetc(f);
    b3 = fgetc(f);
    b4 = fgetc(f);

    return b1 + (b2<<8) + (b3<<16) + (b4<<24);
}
#endif

/*
=======================
32 bit texture loading

this file includes some code from the Independent JPEG Group (IJG).  The original code was
released free of charge to use as one sees fit.  These guys rock.  Check them out at
http://www.ijg.org
=======================
*/
byte *image_rgba;
FILE *file32bit;

int LoadTGA (int *width, int *height, qboolean complain, qboolean blah)
{
#ifdef USE_EXTTEX
    int				w, h, x, y, realrow, truerow, baserow, i, temp1, temp2, pixel_size, map_idx;
    int				RLE_count, RLE_flag, size, interleave, origin;
    qboolean		mapped, rlencoded;
    byte			*dst, r, g, b, a, j, k, l, *ColorMap;
    tga_t		    header;

    header.id_length = fgetc (file32bit);
    header.colormap_type = fgetc (file32bit);
    header.image_type = fgetc (file32bit);
    header.colormap_index = fgetLittleShort (file32bit);
    header.colormap_length = fgetLittleShort (file32bit);
    header.colormap_size = fgetc (file32bit);
    header.x_origin = fgetLittleShort (file32bit);
    header.y_origin = fgetLittleShort (file32bit);
    header.width = fgetLittleShort (file32bit);
    header.height = fgetLittleShort (file32bit);
    header.pixel_size = fgetc (file32bit);
    header.attributes = fgetc (file32bit);

    if (header.id_length != 0)
    {
        fseek(file32bit, header.id_length, SEEK_CUR);  // skip TARGA image comment
    }

    /* validate TGA type */
    switch (header.image_type)
    {
    case TGA_Map:
    case TGA_RGB:
    case TGA_Mono:
    case TGA_RLEMap:
    case TGA_RLERGB:
    case TGA_RLEMono:
        break;

    default:
        Con_SafePrintf ("Unsupported TGA image %s: Only type 1 (map), 2 (RGB), 3 (mono), 9 (RLEmap), 10 (RLERGB), 11 (RLEmono) TGA images supported\n");
        return 0;
    }

    /* validate color depth */
    switch (header.pixel_size)
    {
    case 8:
    case 15:
    case 16:
    case 24:
    case 32:
        break;

    default:
        Con_SafePrintf ("Unsupported TGA image %s: Only 8, 15, 16, 24 or 32 bit images (with colormaps) supported\n");
        return 0;
    }
    r = g = b = a = l = 0;

    /* if required, read the color map information. */
    ColorMap = NULL;
    mapped = (header.image_type == TGA_Map || header.image_type == TGA_RLEMap) && header.colormap_type == 1;

    if (mapped)
    {
        /* validate colormap size */
        switch (header.colormap_size)
        {
        case 8:
        case 15:
        case 16:
        case 32:
        case 24:
            break;

        default:
            Con_SafePrintf ("Unsupported TGA image %s: Only 8, 15, 16, 24 or 32 bit colormaps supported\n");
            return 0;
        }
        temp1 = header.colormap_index;
        temp2 = header.colormap_length;

        if ((temp1 + temp2 + 1) >= TGA_MAXCOLORS)
        {
            return 0;
        }
        ColorMap = malloc (TGA_MAXCOLORS * 4);
        map_idx = 0;

        for (i = temp1 ; i < temp1 + temp2 ; ++i, map_idx += 4)
        {
            /* read appropriate number of bytes, break into rgba & put in map. */
            switch (header.colormap_size)
            {
            case 8:	/* grey scale, read and triplicate. */
                r = g = b = fgetc (file32bit);
                a = 255;
                break;

            case 15:	/* 5 bits each of red green and blue. */
                /* watch byte order. */
                j = fgetc (file32bit);
                k = fgetc (file32bit);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = 255;
                break;

            case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
                /* watch byte order. */
                j = fgetc (file32bit);
                k = fgetc (file32bit);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = (k & 0x80) ? 255 : 0;
                break;

            case 24:	/* 8 bits each of blue, green and red. */
                b = fgetc (file32bit);
                g = fgetc (file32bit);
                r = fgetc (file32bit);
                a = 255;
                l = 0;
                break;

            case 32:	/* 8 bits each of blue, green, red and alpha. */
                b = fgetc (file32bit);
                g = fgetc (file32bit);
                r = fgetc (file32bit);
                a = fgetc (file32bit);
                l = 0;
                break;
            }
            ColorMap[map_idx+0] = r;
            ColorMap[map_idx+1] = g;
            ColorMap[map_idx+2] = b;
            ColorMap[map_idx+3] = a;
        }
    }

    /* check run-length encoding. */
    rlencoded = (header.image_type == TGA_RLEMap || header.image_type == TGA_RLERGB || header.image_type == TGA_RLEMono);
    RLE_count = RLE_flag = 0;

    /* bitchin */
	if (complain)
	{
		if (*width != header.width) return 0;
		if (*height != header.height) return 0;
	}

    // fill in properties
    *width = w = header.width;
    *height = h = header.height;

    size = w * h * 4;
    image_rgba = malloc (size);

    /* read the Targa file body and convert to portable format. */
    pixel_size = header.pixel_size;
    origin = (header.attributes & 0x20) >> 5;
    interleave = (header.attributes & 0xC0) >> 6;
    truerow = baserow = 0;

    for (y=0 ; y<h ; y++)
    {
        realrow = truerow;

        if (origin == TGA_O_UPPER)
        {
            realrow = h - realrow - 1;
        }
        dst = image_rgba + realrow * w * 4;

        for (x=0 ; x<w ; x++)
        {
            /* check if run length encoded. */
            if (rlencoded)
            {
                if (!RLE_count)
                {
                    /* have to restart run. */
                    i = fgetc (file32bit);

                    RLE_flag = (i & 0x80);

                    if (!RLE_flag)	// stream of unencoded pixels
                    {
                        RLE_count = i + 1;
                    }
                    else		// single pixel replicated
                    {
                        RLE_count = i - 127;
                    }

                    /* decrement count & get pixel. */
                    --RLE_count;
                }
                else
                {
                    /* have already read count & (at least) first pixel. */
                    --RLE_count;

                    if (RLE_flag)
                    {
                        /* replicated pixels. */
                        goto PixEncode;
                    }
                }
            }

            /* read appropriate number of bytes, break into RGB. */
            switch (pixel_size)
            {
            case 8:	/* grey scale, read and triplicate. */
                r = g = b = l = fgetc (file32bit);
                a = 255;
                break;

            case 15:	/* 5 bits each of red green and blue. */
                /* watch byte order. */
                j = fgetc (file32bit);
                k = fgetc (file32bit);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = 255;
                break;

            case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
                /* watch byte order. */
                j = fgetc (file32bit);
                k = fgetc (file32bit);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = (k & 0x80) ? 255 : 0;
                break;

            case 24:	/* 8 bits each of blue, green and red. */
                b = fgetc (file32bit);
                g = fgetc (file32bit);
                r = fgetc (file32bit);
                a = 255;
                l = 0;
                break;

            case 32:	/* 8 bits each of blue, green, red and alpha. */
                b = fgetc (file32bit);
                g = fgetc (file32bit);
                r = fgetc (file32bit);
                a = fgetc (file32bit);
                l = 0;
                break;

            default:
                Con_SafePrintf ("Illegal pixel_size: '%d'\n", pixel_size);

                if (mapped)
                {
                    free (ColorMap);
                }
                return 0;
            }

PixEncode:
            if (mapped)
            {
                map_idx = l * 4;
                *dst++ = ColorMap[map_idx+0];
                *dst++ = ColorMap[map_idx+1];
                *dst++ = ColorMap[map_idx+2];
                *dst++ = ColorMap[map_idx+3];
            }
            else
            {
                *dst++ = r;
                *dst++ = g;
                *dst++ = b;
                *dst++ = a;
            }
        }

        if (interleave == TGA_IL_Four)
        {
            truerow += 4;
        }
        else if (interleave == TGA_IL_Two)
        {
            truerow += 2;
        }
        else
        {
            truerow++;
        }

        if (truerow >= h)
        {
            truerow = ++baserow;
        }
    }

    if (mapped)
    {
        free (ColorMap);
    }
    return 1;
#else
    return 0;
#endif
}

int LoadJPG (int *width, int *height, qboolean complain, qboolean blah)
{
#ifdef USE_EXTTEX
#ifdef USE_JPEG
    struct      jpeg_decompress_struct cinfo;
    JDIMENSION  num_scanlines;
    JSAMPARRAY  in;
    struct      jpeg_error_mgr jerr;
    int         numPixels, row_stride;
    byte        *out;
    int          i, r, g, b, count;

    // set up the decompression.
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress (&cinfo);

    // inititalize the source
    jpeg_stdio_src (&cinfo, file32bit);

    // initialize decompression
    (void) jpeg_read_header (&cinfo, TRUE);
    (void) jpeg_start_decompress (&cinfo);

    // set up the width and height for return
    if (complain)
    {
        if (*width != cinfo.image_width) return 0;
        if (*height != cinfo.image_height) return 0;
    }

    *width = cinfo.image_width;
    *height = cinfo.image_height;
    numPixels = (*width) * (*height);

    // initialize the input buffer - we'll use the in-built memory management routines in the
    // JPEG library because it will automatically free the used memory for us when we destroy
    // the decompression structure.  cool.
    row_stride = cinfo.output_width * cinfo.output_components;
    in = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

    // bit of error checking
    if (cinfo.output_components != 3)
        goto error;

    if ((numPixels * 4) != ((row_stride * cinfo.output_height) + numPixels))
        goto error;

    // read the jpeg
    count = 0;

    // initialize the return data
    image_rgba = malloc ((numPixels * 4));

    while (cinfo.output_scanline < cinfo.output_height)
    {
        num_scanlines = jpeg_read_scanlines(&cinfo, in, 1);
        out = in[0];

        for (i = 0; i < row_stride;)
        {
            r = image_rgba[count++] = out[i++];
            g = image_rgba[count++] = out[i++];
            b = image_rgba[count++] = out[i++];
            image_rgba[count++] = 255;
        }
    }

    // finish decompression and destroy the jpeg
    (void) jpeg_finish_decompress (&cinfo);
    jpeg_destroy_decompress (&cinfo);

    return 1;

error:
    // this should rarely (if ever) happen, but just in case...
    Con_DPrintf ("Invalid JPEG Format\n");
    jpeg_destroy_decompress (&cinfo);
#endif
#endif

    return 0;
}


int LoadPCX (int *width, int *height, qboolean complain, qboolean use_fake_alpha)
{
#ifdef USE_EXTTEX
    pcx_t	*pcx, pcxbuf;
    byte	palette[768];
    byte	*pix;
    int		x, y;
    int		dataByte, runLength;
    int		count;

    // parse the PCX file
    fread (&pcxbuf, 1, sizeof(pcxbuf), file32bit);

    pcx = &pcxbuf;

    if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 320 || pcx->ymax >= 256)
    {
        Con_DPrintf ("Bad pcx file\n");
        return 0;
    }

    if (complain)
    {
        if (*width != pcx->xmax) return 0;
        if (*width != pcx->ymax) return 0;
    }

    // we have to use +1 here cos otherwise pcx textures will be slanted.  this is cos pcx
    // image dimensions start at 0 (eg. 0 to 127) whereas quake textures start at 1 (1 to 128)
    *width = pcx->xmax + 1;
    *height = pcx->ymax + 1;

    // seek to palette
    fseek (file32bit, -768, SEEK_END);
    fread (palette, 1, 768, file32bit);

    fseek (file32bit, sizeof(pcxbuf) - 4, SEEK_SET);

    count = (pcx->xmax+1) * (pcx->ymax+1);
    image_rgba = malloc( count * 4);

    for (y = 0; y <= pcx->ymax; y++)
    {
        pix = image_rgba + 4*y*(pcx->xmax+1);

        for (x = 0; x <= pcx->xmax; )	// was ymax!!!
        {
            dataByte = fgetc(file32bit);

            if((dataByte & 0xC0) == 0xC0)
            {
                runLength = dataByte & 0x3F;
                dataByte = fgetc(file32bit);
            }
            else runLength = 1;

            while (runLength-- > 0)
            {
                pix[0] = palette[dataByte*3];
                pix[1] = palette[dataByte*3+1];
                pix[2] = palette[dataByte*3+2];
                pix[3] = 255;

                if (pix[0] == 159 && pix[1] == 91 && pix[2] == 83 && use_fake_alpha)
                {
                    // 'fake' alpha for PCX textures - keep the classic quake alpha colour.
                    pix[0] = 0;
                    pix[1] = 0;
                    pix[2] = 0;
                    pix[3] = 0;
                }

                pix += 4;
                x++;
            }
        }
    }

    return 1;
#else
    return 0;
#endif
}


#ifdef USE_EXTTEX
// got rid of a lot of duplicated code here...
int CheckTexImage (char *ext, char *new_ident, char *loadname, int *width, int *height, qboolean mipmap, qboolean alpha, int (*LoadFunc) (), qboolean fake_pcx_alpha)
{
    int     i, j;
    char    full_path[1024];
    int     return_rgba;
    int     old_width, old_height;

    old_width = *width;
    old_height = *height;

    sprintf (full_path, "textures/%s.%s", new_ident, ext);
    i = COM_FOpenFile (full_path, &file32bit);

    if (file32bit)
    {
        // using pointers here will give us the correct width and height of the external texture
        j = (*LoadFunc) (width, height, false, fake_pcx_alpha);

        if (j)
        {
            // we can't return straight away because we need to clean up after ourselves first
            if (!gl_texscale.value)
            {
                return_rgba = MHGL_LoadTexture (loadname, old_width, old_height, image_rgba, mipmap, alpha, 32);
            }
            else
            {
                return_rgba = MHGL_LoadTexture (loadname, *width, *height, image_rgba, mipmap, alpha, 32);
            }

            // even if we have an invalid image, the file is still open (although image_rgba won't
            // have been malloced yet) so we need to close it.
            free (image_rgba);
            fclose (file32bit);

            // because some of the loading functions can create an error after modifying height
            // and width, we must reset them for the actual load of the internal texture.
            *width = old_width;
            *height = old_height;
        }

        // oups we returned the value before resetting
        // the size :( making textures scale wrong. moved here.
        return return_rgba;
    }
    return -1;
}
#endif


int GL_LoadTexImage (char *identifier, char *old_ident, int *width, int *height, byte *data, qboolean mipmap, qboolean alpha, qboolean cache)
{
    char    loadname[64];
#ifdef USE_EXTTEX
    char    new_ident[64];
    int     return_rgba;
    int     i, j;
#endif

    // identify water texture
    if (identifier[0] == '*')
         is_water = true;
    else is_water = false;

#ifdef USE_EXTTEX
    // strip invalid characters from the identifier
    for (i = 0, j = 0; i < strlen (identifier); i++)
    {
        // remove the extension if it has one.
        if (identifier[i] == '.')
        {
            i += 3;
            if (i >= strlen (identifier)) break;
            continue;
        }

        // replace water identifiers
        if (identifier[i] == '*')
        {
            identifier[i] = '#';
        }

        // don't remove slashes cos they indicate subdirectories of the textures directory
        if (identifier[i] == '*' || identifier[i] == ':' ||
            identifier[i] == '?' || identifier[i] == '"' ||
            identifier[i] == '<' || identifier[i] == ' ' ||
            identifier[i] == '>' || identifier[i] == '|') continue;

        new_ident[j] = identifier[i];
        j++;
    }
    new_ident[j] = '\0';

    // testing - we only want this if developer is 2
    Con_D2Printf ("Safe Texture Name: %s\n", new_ident);
#endif

    // if we're going to be replacing this texture, then make loadname spurious, otherwise make
    // it equal to identifier.
    if (!cache)
        strcpy (loadname, "Jesus loves MHQuake");
    else strcpy (loadname, old_ident);

#ifdef USE_EXTTEX
    // look for a TGA
    return_rgba = CheckTexImage ("tga", new_ident, loadname, width, height, mipmap, alpha, LoadTGA, false);
    if (return_rgba != -1) return return_rgba;

    // look for a JPG - no alpha here
    return_rgba = CheckTexImage ("jpg", new_ident, loadname, width, height, mipmap, false, LoadJPG, false);
    if (return_rgba != -1) return return_rgba;

    // look for a PCX
    return_rgba = CheckTexImage ("pcx", new_ident, loadname, width, height, mipmap, alpha, LoadPCX, (!mipmap && alpha));
    if (return_rgba != -1) return return_rgba;
#endif

    // use NULL as the data to attempt to load just a TGA, a JPG or a PCX and have your own
    // default behaviour defined in the function you call this from...
    if (!data) return -1;

    // default back to the parameters it was called with
    return MHGL_LoadTexture (loadname, *width, *height, data, mipmap, true, 8);
}


int GL_LoadTexture (char *identifier, int *width, int *height, byte *data, qboolean mipmap, qboolean alpha)
{
    // load a permanent texture
    return GL_LoadTexImage (identifier, identifier, width, height, data, mipmap, alpha, true);
}


int GL_LoadTextureNoCache (char *identifier)
{
    int i = 0;

    // load a texture into the cache - it can be replaced the next time this function is called
    return GL_LoadTexImage (identifier, identifier, &i, &i, NULL, false, false, false);
}


/*
================
GL_LoadPicTexture
================
*/
int GL_LoadPicTexture (qpic_t *pic, char *pic_name)
{
    // load a texture into the cache - in theory, these can be replaced, in practice, they are
    // needed so often that it's better not to...
    return GL_LoadTexImage (pic_name, pic_name, &pic->width, &pic->height, pic->data, false, true, true);
}


void Load_Conback (void);

void GL_LoadTextureReplace (char *texname, char *oldname, qboolean report)
{
    gl_texture_t *oldtex;
    int newtexnum;
    int texhack = 64;

    // first off, see if the texture exists to be replaced
    oldtex = GL_FindFullTexture (oldname);

    if (!oldtex)
    {
        if (report) Con_Printf ("%s is not loaded\n", oldname);
        return;
    }

    // using the original mipmap and alpha causes the texture to go corrupt, so we won't do it
    // until we find out why...
    newtexnum = GL_LoadTexImage (texname, oldname, &texhack, &texhack, NULL, false, false, true);

    if (newtexnum == -1)
    {
        if (report) Con_Printf ("Replace failed - cannot find %s\n", texname);
        if (!strcmp (oldname, "conback")) Load_Conback ();
        return;
    }

    if (report) Con_Printf ("Replacing %s with %s\n", oldname, texname);
}

