/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// brush model loading functions - seperated out to make gl_model.c more readable

#include "quakedef.h"
#include "gl_mtex.h"
#include "gl_mirror.h"

byte	*mod_base;

int		global_worldtype;
int		worldtype_found;
int		use_lit;

qboolean hlbsp;
qboolean qsg30;

/*
=================
Load Textures
=================
*/
extern int fog_colour[3];
extern int fog_density;
extern cvar_t _r_skybox;
void Host_LoadSkybox (char *sb_name, qboolean report);


void Mod_LoadTextures (lump_t *l)
{
    int		i, j, k, pixels, num, max, altmax;
    miptex_t	*mt;
    gltexture_t	*tx, *tx2;
    gltexture_t	*anims[10];
    gltexture_t	*altanims[10];
    dmiptexlump_t *m;
    char tex_name[64];
    qboolean mipmap;

    if (!l->filelen)
    {
        loadmodel->textures = NULL;

        // quake is gonna crash here anyway...
        return;
    }

    m = (dmiptexlump_t *)(mod_base + l->fileofs);
    m->nummiptex = LittleLong (m->nummiptex);

    loadmodel->numtextures = m->nummiptex;
    loadmodel->textures = Hunk_AllocName (m->nummiptex * sizeof(*loadmodel->textures) , loadname);

    for (i = 0, k = 0; i < m->nummiptex; i++)
    {
        m->dataofs[i] = LittleLong(m->dataofs[i]);

        if (m->dataofs[i] == -1)
            continue;

        mt = (miptex_t *)((byte *)m + m->dataofs[i]);
        mt->width = LittleLong (mt->width);
        mt->height = LittleLong (mt->height);

        for (j=0 ; j<MIPLEVELS ; j++)
            mt->offsets[j] = LittleLong (mt->offsets[j]);

        if ((mt->width & 15) || (mt->height & 15))
            Sys_Error ("Texture %s is not 16 aligned", mt->name);

        pixels = mt->width*mt->height/64*85;

        tx = Hunk_AllocName (sizeof(gltexture_t) + pixels, loadname);
        loadmodel->textures[i] = tx;

        memcpy (tx->name, mt->name, sizeof(tx->name));
        tx->width = mt->width;
        tx->height = mt->height;

#ifdef HALFLIFEBSP
        if (hlbsp)
        {
            // halflife textures are followed by a short (no. of palette entries) and then
            // 256 RGB triplets
            pixels += 770;
            Mod_LoadHLTextures (mt, tx, pixels);
            tx->fullbright = -1;
        }
        else
#endif
        {
            for (j=0 ; j<MIPLEVELS ; j++)
                tx->offsets[j] = mt->offsets[j] + sizeof(gltexture_t) - sizeof(miptex_t);

            // the pixels immediately follow the structures
            memcpy ( tx+1, mt+1, pixels);

            // halflife has textures called 'sky' but they are replaced with skyboxes at runtime
            if (!strncmp(mt->name,"sky",3))
                R_InitSky (tx);
            else
            {
                fog_colour[0] = fog_colour[1] = fog_colour[2] = fog_density = 255;

                texture_mode = GL_LINEAR_MIPMAP_LINEAR;

                // don't mipmap alpha-test textures
                if (mt->name[0] == '{')
                    mipmap = false;
                else mipmap = true;

                // assume alpha is true on all textures - gl_upload8 will discard alpha on all
                // textures that have no alpha components anyway
                tx->gl_texturenum = GL_LoadTexture (mt->name, &tx->width, &tx->height,
                                                    (byte *)(tx+1), mipmap, true);

                tx->water_fog_density = (float) fog_density / 255.0;
                tx->water_fog_colour[0] = (float) fog_colour[0] / 255.0;
                tx->water_fog_colour[1] = (float) fog_colour[1] / 255.0;
                tx->water_fog_colour[2] = (float) fog_colour[2] / 255.0;

                // check for fullbright pixels in the texture - only if it ain't liquid, etc also
                if ((tx->name[0] != '*' && tx->name[0] != '~' && tx->name[0] != '!') &&
                        (Mod_FindFullBrights ((byte *)(tx+1), pixels)))
                {
                    // convert any non fullbright pixel to fully transparent
                    Mod_ConvertFullbrights ((byte *)(tx + 1), pixels);

                    // get a new name for the fullbright mask to avoid cache mismatches
                    sprintf (tex_name, "fbrm_%s", mt->name);

                    // load the fullbright pixels version of the texture
                    tx->fullbright = GL_LoadTexture (tex_name, &tx->width, &tx->height,
                                                     (byte *)(tx + 1), true, true);
                }
                else tx->fullbright = -1;	// because 0 is a potentially valid texture number

                texture_mode = GL_LINEAR;
            }
        }
    }

    // sequence the animations
    for (i=0 ; i<m->nummiptex ; i++)
    {
        tx = loadmodel->textures[i];

        if (!tx || tx->name[0] != '+')
            continue;

        if (tx->anim_next)
            continue;	// allready sequenced

        // find the number of frames in the animation
        memset (anims, 0, sizeof(anims));
        memset (altanims, 0, sizeof(altanims));

        max = tx->name[1];
        altmax = 0;

        if (max >= 'a' && max <= 'z')
            max -= 'a' - 'A';

        if (max >= '0' && max <= '9')
        {
            max -= '0';
            altmax = 0;
            anims[max] = tx;
            max++;
        }
        else if (max >= 'A' && max <= 'J')
        {
            altmax = max - 'A';
            max = 0;
            altanims[altmax] = tx;
            altmax++;
        }
        else
            Sys_Error ("Bad animating texture %s", tx->name);

        for (j=i+1 ; j<m->nummiptex ; j++)
        {
            tx2 = loadmodel->textures[j];

            if (!tx2 || tx2->name[0] != '+')
                continue;

            if (strcmp (tx2->name+2, tx->name+2))
                continue;

            num = tx2->name[1];
            if (num >= 'a' && num <= 'z')
                num -= 'a' - 'A';

            if (num >= '0' && num <= '9')
            {
                num -= '0';
                anims[num] = tx2;
                if (num+1 > max)
                    max = num + 1;
            }
            else if (num >= 'A' && num <= 'J')
            {
                num = num - 'A';
                altanims[num] = tx2;
                if (num+1 > altmax)
                    altmax = num+1;
            }
            else
                Sys_Error ("Bad animating texture %s", tx->name);
        }

#define	ANIM_CYCLE	2
        // link them all together
        for (j=0 ; j<max ; j++)
        {
            tx2 = anims[j];

            if (!tx2)
                Sys_Error ("Missing frame %i of %s",j, tx->name);

            tx2->anim_total = max * ANIM_CYCLE;
            tx2->anim_min = j * ANIM_CYCLE;
            tx2->anim_max = (j+1) * ANIM_CYCLE;
            tx2->anim_next = anims[ (j+1)%max ];

            if (altmax)
                tx2->alternate_anims = altanims[0];
        }

        for (j=0 ; j<altmax ; j++)
        {
            tx2 = altanims[j];

            if (!tx2)
                Sys_Error ("Missing frame %i of %s",j, tx->name);

            tx2->anim_total = altmax * ANIM_CYCLE;
            tx2->anim_min = j * ANIM_CYCLE;
            tx2->anim_max = (j+1) * ANIM_CYCLE;
            tx2->anim_next = altanims[ (j+1)%altmax ];

            if (max)
                tx2->alternate_anims = anims[0];
        }
    }

    // load the skybox texture (if there is one...) - don't report errors
    Host_LoadSkybox (_r_skybox.string, false);
}


/*
=================
Mod_LoadLighting
=================
*/
// litfile header - darkplaces standard
typedef struct litheader_s
{
    char ident[4];
    int version;
} litheader_t;

// support for bsp29, bsp30 and lit files
// revised - only tries to load a lit file if (a) we have no lightdata in the bsp, or
// (b) we are using coloured light - otherwise we don't need to!!!!
// revised - if a valid litfile is present for a qsg 30 BSP the light offsets in the faces
// will be all wrong - in any event we don't need a litfile for a qsg 30 BSP so we don't
// even bother loading one.
void Mod_LoadLighting (lump_t *l)
{
    int			i, j;
    int			r, g, b;
    byte		*data;
	FILE		*litfile;
    char		litfilename[1024];
    litheader_t litfileheader;
    byte		*explit;

#ifdef HALFLIFEBSP
    if (hlbsp)
    {
        Mod_LoadHLLighting (l);
        return;
    }
#endif

    loadmodel->lightdata = NULL;

    // if it's a version 30 QSG BSP then don't even bother trying for a LIT file!
    if (qsg30) goto LoadQSG30Light;

    if (l->filelen == 0 || gl_useclight)
    {
        strcpy(litfilename, loadmodel->name);
        COM_StripExtension(litfilename, litfilename);
        strcat(litfilename, ".lit");

        COM_FOpenFile (litfilename, &litfile);
        use_lit = 0;

        if (litfile)
        {
            fread (&litfileheader, sizeof (litfileheader), 1, litfile);

            if (!strncmp (litfileheader.ident, "QLIT", 4))
            {
                if (litfileheader.version == 1)
                {
                    // load and condense the litfile
                    Con_DPrintf ("Using Light Data from %s - DarkPlaces format\n", litfilename);
                    use_lit = 1;

                    data = (byte *) malloc (l->filelen * 3);
                    fread (data, l->filelen * 3, 1, litfile);

                    // because lit files can also be used by versions of light.exe that spew
                    // out a lit file directly rather than writing it to the bsp, we need to
                    // retain the option to use coloured light or not here
                    // much later: what the fuck am i trying to say here????
                    if (gl_useclight)
                    {
                        // expand RGB to RGBA
                        explit = (byte *) malloc (l->filelen * 4);

                        for (i = 0, j = 0; i < l->filelen * 3; i += 3, j += 4)
                        {
                            r = explit[j] = data[i];
                            g = explit[j+1] = data[i+1];
                            b = explit[j+2] = data[i+2];
                            explit[j+3] = (r + g + b) / 3;
                        }

                        free (data);
                        loadmodel->lightdata = Hunk_AllocName (l->filelen * 4, litfilename);
                        memcpy (loadmodel->lightdata, explit, l->filelen * 4);
                    }
                    else
                    {
                        // condense to luminance
                        explit = (byte *) malloc (l->filelen);

                        for (i = 0, j = 0; i < l->filelen * 3; i += 3, j++)
                        {
                            r = data[i];
                            g = data[i+1];
                            b = data[i+2];
                            explit[j] = (r + g + b) / 3;
                        }
                        free (data);
                        loadmodel->lightdata = Hunk_AllocName (l->filelen, litfilename);
                        memcpy (loadmodel->lightdata, explit, l->filelen);
                    }
                    free (explit);
                    fclose (litfile);

                    return;
                }
                else Con_DPrintf("Unknown .lit file version (%d)\n", litfileheader.version);
            }
            else Con_DPrintf("Corrupt .lit file (old version?), ignoring\n");

            fclose (litfile);
        }
    }

LoadQSG30Light:;
    // any error in the lit file will cause the engine to load the lightdata from the BSP
    if (!l->filelen) return;

    // qsg version 30 bsp format
    if (qsg30)
    {
        Con_DPrintf ("Using Light Data from %s - BSP 30 format\n", loadmodel->name);

        if (gl_useclight)
        {
            // load from the bsp
            loadmodel->lightdata = Hunk_AllocName ( l->filelen, loadname);
            memcpy (loadmodel->lightdata, mod_base + l->fileofs, l->filelen);
        }
        else
        {
            // load from the bsp and condense
            loadmodel->lightdata = Hunk_AllocName (l->filelen / 4, loadname);
            explit = (byte *) malloc (l->filelen);
            memcpy (explit, mod_base + l->fileofs, l->filelen);

            for (i = 0, j = 0; i < l->filelen; i += 4, j++)
            {
                r = explit[i];
                g = explit[i+1];
                b = explit[i+2];
                loadmodel->lightdata[j] = (r + g + b) / 3;
            }
            free (explit);
        }
        return;
    }

    // bsp29
    Con_DPrintf ("Using Light Data from %s - BSP 29 format\n", loadmodel->name);

    if (gl_useclight)
    {
        // expand to rgba
        loadmodel->lightdata = Hunk_AllocName (l->filelen * 4, loadname);
        explit = (byte *) malloc (l->filelen);
        memcpy (explit, mod_base + l->fileofs, l->filelen);

        for (i = 0, j = 0; i < l->filelen; i++, j += 4)
        {
            loadmodel->lightdata[j] = explit[i];
            loadmodel->lightdata[j+1] = explit[i];
            loadmodel->lightdata[j+2] = explit[i];
            loadmodel->lightdata[j+3] = explit[i];
        }
        free (explit);
    }
    else
    {
        // load as normal
        loadmodel->lightdata = Hunk_AllocName ( l->filelen, loadname);
        memcpy (loadmodel->lightdata, mod_base + l->fileofs, l->filelen);
    }
}

/*
=================
Mod_LoadVisibility
=================
*/
void Mod_LoadVisibility (lump_t *l)
{
    if (!l->filelen)
    {
        loadmodel->visdata = NULL;
        return;
    }

    loadmodel->visdata = Hunk_AllocName ( l->filelen, loadname);
    memcpy (loadmodel->visdata, mod_base + l->fileofs, l->filelen);
}

/*
=================
Mod_LoadVertexes
=================
*/
void Mod_LoadVertexes (lump_t *l)
{
    dvertex_t	*in;
    mvertex_t	*out;
    int			i, count;

    in = (void *)(mod_base + l->fileofs);
    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->vertexes = out;
    loadmodel->numvertexes = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        out->position[0] = LittleFloat (in->point[0]);
        out->position[1] = LittleFloat (in->point[1]);
        out->position[2] = LittleFloat (in->point[2]);
    }
}

/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l)
{
    dmodel_t	*in;
    dmodel_t	*out;
    int			i, j, count;

    in = (void *)(mod_base + l->fileofs);
    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName (count*sizeof(*out), loadname);

    loadmodel->submodels = out;
    loadmodel->numsubmodels = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j=0 ; j<3 ; j++)
        {
            // spread the mins / maxs by a pixel
            out->mins[j] = LittleFloat (in->mins[j]) - 1;
            out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
            out->origin[j] = LittleFloat (in->origin[j]);
        }

        for (j=0 ; j<MAX_MAP_HULLS ; j++)
            out->headnode[j] = LittleLong (in->headnode[j]);

        out->visleafs = LittleLong (in->visleafs);
        out->firstface = LittleLong (in->firstface);
        out->numfaces = LittleLong (in->numfaces);
    }
}

/*
=================
Mod_LoadEdges
=================
*/
void Mod_LoadEdges (lump_t *l)
{
    dedge_t *in;
    medge_t *out;
    int 	i, count;

    in = (void *)(mod_base + l->fileofs);

    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( (count + 1) * sizeof(*out), loadname);

    loadmodel->edges = out;
    loadmodel->numedges = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        out->v[0] = (unsigned short)LittleShort(in->v[0]);
        out->v[1] = (unsigned short)LittleShort(in->v[1]);
    }
}

/*
=================
Mod_LoadTexinfo
=================
*/
void Mod_LoadTexinfo (lump_t *l)
{
    texinfo_t *in;
    glmtexinfo_t *out;
    int 	i, j, count;
    int		miptex;
    float	len1, len2;

    in = (void *)(mod_base + l->fileofs);

    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->texinfo = out;
    loadmodel->numtexinfo = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j=0 ; j<8 ; j++)
            out->vecs[0][j] = LittleFloat (in->vecs[0][j]);

        len1 = VLength (out->vecs[0]);
        len2 = VLength (out->vecs[1]);
        len1 = (len1 + len2)/2;

        if (len1 < 0.32)
            out->mipadjust = 4;
        else if (len1 < 0.49)
            out->mipadjust = 3;
        else if (len1 < 0.99)
            out->mipadjust = 2;
        else
            out->mipadjust = 1;

        miptex = LittleLong (in->miptex);
        out->flags = LittleLong (in->flags);

        if (!loadmodel->textures)
        {
            out->texture = r_notexture_mip;	// checkerboard texture
            out->flags = 0;
        }
        else
        {
            if (miptex >= loadmodel->numtextures)
                Sys_Error ("miptex >= loadmodel->numtextures");

            out->texture = loadmodel->textures[miptex];

            if (!out->texture)
            {
                out->texture = r_notexture_mip; // texture not found
                out->flags = 0;
            }
        }
    }
}

/*
================
CalcSurfaceExtents

Fills in s->texturemins[] and s->extents[]
================
*/
void CalcSurfaceExtents (glmsurface_t *s)
{
    float	mins[2], maxs[2], val;
    int		i,j, e;
    mvertex_t	*v;
    glmtexinfo_t	*tex;
    int		bmins[2], bmaxs[2];

    mins[0] = mins[1] = 999999;
    maxs[0] = maxs[1] = -99999;

    tex = s->texinfo;

    for (i=0 ; i<s->numedges ; i++)
    {
        e = loadmodel->surfedges[s->firstedge+i];

        if (e >= 0)
            v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
        else v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];

        for (j=0 ; j<2 ; j++)
        {
            val = v->position[0] * tex->vecs[j][0] +
                  v->position[1] * tex->vecs[j][1] +
                  v->position[2] * tex->vecs[j][2] +
                  tex->vecs[j][3];

            if (val < mins[j]) mins[j] = val;
            if (val > maxs[j]) maxs[j] = val;
        }
    }

    s->BadExtents = false;

    for (i=0 ; i<2 ; i++)
    {
        bmins[i] = floor(mins[i]/16);
        bmaxs[i] = ceil(maxs[i]/16);

        s->texturemins[i] = bmins[i] * 16;
        s->extents[i] = (bmaxs[i] - bmins[i]) * 16;

        // baaad - when this happens, we've got a qbsp problem where a regular surface
        // (not a sky or liquid one) has been made too large.  it's ok for sky/liquid to
        // be large cos they'll subsequently be subdivided but not for solids.  since
        // extents are just used for lightmaps (each surf has it's own lightmap texture
        // which is a subimage of a larger texture - hence the size limit) we will flag the
        // surface as having bad extents instead and render it fullbright when the time comes.
        if ( !(tex->flags & TEX_SPECIAL) && s->extents[i] > 512)
		{
            s->BadExtents = true;
		}
    }
}


/*
=================
Mod_LoadFaces
=================
*/
void Mod_LoadFaces (lump_t *l)
{
    dface_t		*in;
    glmsurface_t 	*out;
    int			i, count, surfnum;
    int			planenum, side;

    in = (void *)(mod_base + l->fileofs);

    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->surfaces = out;
    loadmodel->numsurfaces = count;

    for (surfnum = 0; surfnum < count; surfnum++, in++, out++)
    {
        out->firstedge = LittleLong(in->firstedge);
        out->numedges = LittleShort(in->numedges);
        out->flags = 0;

        planenum = LittleShort(in->planenum);
        side = LittleShort(in->side);

        if (side) out->flags |= SURF_PLANEBACK;

        out->plane = loadmodel->planes + planenum;
        out->texinfo = loadmodel->texinfo + LittleShort (in->texinfo);
        CalcSurfaceExtents (out);

        // lighting info
        for (i=0 ; i<MAXLIGHTMAPS ; i++)
            out->styles[i] = in->styles[i];

        i = LittleLong(in->lightofs);

        if (i == -1)
            out->samples = NULL;
#ifdef HALFLIFEBSP
        else if (hlbsp)
        {
            if (gl_useclight)
                out->samples = loadmodel->lightdata + (i / 3) * 4;
            else out->samples = loadmodel->lightdata + (i / 3);
        }
#endif
        else if (qsg30)
        {
            if (gl_useclight)
                out->samples = loadmodel->lightdata + i;
            else out->samples = loadmodel->lightdata + (i / 4);
        }
        else
        {
            if (gl_useclight)
                out->samples = loadmodel->lightdata + (i * 4);
            else out->samples = loadmodel->lightdata + i;
        }

        // initialize at NULL to prevent any pointer problems later on
        //out->polys = NULL;
        //out->warp_polys = NULL;

        // set the drawing flags flag
        if (!strncmp(out->texinfo->texture->name,"sky",3))	// sky
        {
            out->flags |= (SURF_DRAWSKY | SURF_DRAWTILED);

            // don't subdivide halflife skies
            if (!hlbsp)
            {
                GL_SubdivideSurface (out, 256);	// cut up polygon for warps
            }
            continue;
        }

        if (out->texinfo->texture->name[0] == '*' && !hlbsp)		// turbulent
        {
            if (out->texinfo->texture->name[1] == '*')
                out->flags |= (SURF_DRAWTURB|SURF_DRAWTILED|SURF_DRAWPERV);
            else out->flags |= (SURF_DRAWTURB|SURF_DRAWTILED);

            // SURF_LAVA, SURF_SLIME and SURF_TELE are only relevant for Quake maps - HL
            // does things quite differently...
            if (!strncmp (out->texinfo->texture->name, "*lava", 5))
                out->flags |= SURF_LAVA;
            else if (!strncmp (out->texinfo->texture->name, "*tele", 5) ||
                     !strncmp (out->texinfo->texture->name, "*rift", 5))
                out->flags |= SURF_TELE;
            else if (!strncmp (out->texinfo->texture->name, "*slime", 6))
                out->flags |= SURF_SLIME;

            for (i=0 ; i<2 ; i++)
            {
                out->extents[i] = 16384;
                out->texturemins[i] = -8192;
            }

            GL_SubdivideSurface (out, 64);	// cut up polygon for warps
            continue;
        }

#ifdef HALFLIFEBSP
        if (!strncmp(out->texinfo->texture->name,"!",1) && hlbsp)		// turbulent
        {
            out->flags |= (SURF_DRAWTURB|SURF_DRAWTILED);

            // do these subdivide in hl?
            for (i=0 ; i<2 ; i++)
            {
                out->extents[i] = 16384;
                out->texturemins[i] = -8192;
            }

            GL_SubdivideSurface (out, 64);	// cut up polygon for warps
            continue;
        }

        // halflife windows
        if (!strncmp (out->texinfo->texture->name, "translucent", 11) && hlbsp)
        {
            out->flags |= SURF_WINDOW;
            continue;
        }

        if (out->texinfo->texture->name[0] == '@' && hlbsp)
        {
            out->flags |= SURF_WINDOW;
            continue;
        }
#endif

        if (out->texinfo->texture->name[0] == '!' && !hlbsp)
        {
            out->flags |= SURF_WINDOW;
            continue;
        }

        // textures starting with '~' are something else in hl that the sdk isn't
        // about to give away... - so just make them normal textures for now
        if (out->texinfo->texture->name[0] == '~' && !hlbsp)
        {
            out->flags |= SURF_FOG;
            continue;
        }

        if (out->texinfo->texture->name[0] == 'f' &&
                out->texinfo->texture->name[1] == 'o' &&
                out->texinfo->texture->name[2] == 'g' && !hlbsp)
        {
            out->flags |= SURF_FOG;
            continue;
        }

        if (out->texinfo->texture->name[0] == '{')
        {
            out->flags |= SURF_TRANS;
            continue;
        }

        if (!strncmp (out->texinfo->texture->name, "window02_1", 10) && !hlbsp)
        {
            out->flags |= SURF_MIRROR;
            continue;
        }

        // might as well be this as anything else...
        if (out->texinfo->texture->name[0] == '&' && !hlbsp)
        {
            out->flags |= SURF_MIRROR;
            continue;
        }
    }
}


/*
=================
Mod_SetParent
=================
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
    node->parent = parent;

    if (node->contents < 0)
        return;

    Mod_SetParent (node->children[0], node);
    Mod_SetParent (node->children[1], node);
}

/*
=================
Mod_LoadNodes
=================
*/
void Mod_LoadNodes (lump_t *l)
{
    int			i, j, count, p;
    dnode_t		*in;
    mnode_t 	*out;

    in = (void *)(mod_base + l->fileofs);

    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->nodes = out;
    loadmodel->numnodes = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j=0 ; j<3 ; j++)
        {
            out->minmaxs[j] = LittleShort (in->mins[j]);
            out->minmaxs[3+j] = LittleShort (in->maxs[j]);
        }

        p = LittleLong(in->planenum);
        out->plane = loadmodel->planes + p;

        out->firstsurface = LittleShort (in->firstface);
        out->numsurfaces = LittleShort (in->numfaces);

        for (j=0 ; j<2 ; j++)
        {
            p = LittleShort (in->children[j]);
            if (p >= 0)
                out->children[j] = loadmodel->nodes + p;
            else
                out->children[j] = (mnode_t *)(loadmodel->leafs + (-1 - p));
        }
    }

    Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}

/*
=================
Mod_LoadLeafs
=================
*/
void Mod_LoadLeafs (lump_t *l)
{
    dleaf_t 	*in;
    mleaf_t 	*out;
    int			i, j, count, p;

    in = (void *)(mod_base + l->fileofs);
    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->leafs = out;
    loadmodel->numleafs = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j=0 ; j<3 ; j++)
        {
            out->minmaxs[j] = LittleShort (in->mins[j]);
            out->minmaxs[3+j] = LittleShort (in->maxs[j]);
        }

        p = LittleLong(in->contents);
        out->contents = p;

        out->firstmarksurface = loadmodel->marksurfaces +
                                LittleShort(in->firstmarksurface);
        out->nummarksurfaces = LittleShort(in->nummarksurfaces);

        p = LittleLong(in->visofs);
        if (p == -1)
            out->compressed_vis = NULL;
        else
            out->compressed_vis = loadmodel->visdata + p;
        out->efrags = NULL;

        for (j=0 ; j<4 ; j++)
            out->ambient_sound_level[j] = in->ambient_level[j];

        // gl underwater warp
        if (out->contents != CONTENTS_EMPTY)
        {
            for (j=0 ; j<out->nummarksurfaces ; j++)
                out->firstmarksurface[j]->flags |= SURF_UNDERWATER;
        }
    }
}

/*
=================
Mod_LoadClipnodes
=================
*/
void Mod_LoadClipnodes (lump_t *l)
{
    dclipnode_t *in, *out;
    int			i, count;
    hull_t		*hull;

    in = (void *)(mod_base + l->fileofs);

    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->clipnodes = out;
    loadmodel->numclipnodes = count;

    hull = &loadmodel->hulls[1];
    hull->clipnodes = out;
    hull->firstclipnode = 0;
    hull->lastclipnode = count-1;
    hull->planes = loadmodel->planes;

    hull->clip_mins[0] = -16;
    hull->clip_mins[1] = -16;
#ifdef HALFLIFEBSP
    if (hlbsp) hull->clip_mins[2] = -36;
    else hull->clip_mins[2] = -24;
#endif

    hull->clip_maxs[0] = 16;
    hull->clip_maxs[1] = 16;
#ifdef HALFLIFEBSP
    if (hlbsp) hull->clip_maxs[2] = 36;
    else hull->clip_maxs[2] = 32;
#endif

    hull = &loadmodel->hulls[2];
    hull->clipnodes = out;
    hull->firstclipnode = 0;
    hull->lastclipnode = count-1;
    hull->planes = loadmodel->planes;

    hull->clip_mins[0] = -32;
    hull->clip_mins[1] = -32;
#ifdef HALFLIFEBSP
    if (hlbsp) hull->clip_mins[2] = -32;
    else hull->clip_mins[2] = -24;
#endif

    hull->clip_maxs[0] = 32;
    hull->clip_maxs[1] = 32;
#ifdef HALFLIFEBSP
    if (hlbsp) hull->clip_maxs[2] = 32;
    else hull->clip_maxs[2] = 64;

    if (hlbsp)
    {
        hull = &loadmodel->hulls[3];
        hull->clipnodes = out;
        hull->firstclipnode = 0;
        hull->lastclipnode = count-1;
        hull->planes = loadmodel->planes;

        hull->clip_mins[0] = -16;
        hull->clip_mins[1] = -16;
        hull->clip_mins[2] = -18;

        hull->clip_maxs[0] = 16;
        hull->clip_maxs[1] = 16;
        hull->clip_maxs[2] = 18;
    }
#endif

    for (i=0 ; i<count ; i++, out++, in++)
    {
        out->planenum = LittleLong(in->planenum);
        out->children[0] = LittleShort(in->children[0]);
        out->children[1] = LittleShort(in->children[1]);
    }
}


/*
=================
Mod_MakeHull0

Duplicate the drawing hull structure as a clipping hull
MH - includes dirty hack to fix sky nodes not being set to CONTENTS_SKY in ID maps
UPDATE - i told you it was a dirty hack - subsequently removed cos it really fucked with
various other nodes, thanks to id's crappy early qbsp version.
=================
*/
void Mod_MakeHull0 (void)
{
    mnode_t		*in, *child;
    dclipnode_t *out;
    int			i, j, count;
    hull_t		*hull;

    hull = &loadmodel->hulls[0];

    in = loadmodel->nodes;
    count = loadmodel->numnodes;
    out = Hunk_AllocName (count * sizeof (*out), loadname);

    hull->clipnodes = out;
    hull->firstclipnode = 0;
    hull->lastclipnode = count-1;
    hull->planes = loadmodel->planes;

    for (i = 0; i < count; i++, out++, in++)
    {
        out->planenum = in->plane - loadmodel->planes;

        for (j = 0; j < 2; j++)
        {
            child = in->children[j];

            if (child->contents < 0)
            {
#ifdef LETS_RUIN_EVERYTHING
                // check the node for a sky texture - this way is much faster than using strncmp
                // we need only check the first surface (if that's sky, the rest of it will be
                // too), and since everything else has been loaded at this stage, it's safe to
                // check it this way.
                if (loadmodel->surfaces[in->firstsurface].texinfo->texture->name[0] == 's' &&
                    loadmodel->surfaces[in->firstsurface].texinfo->texture->name[1] == 'k' &&
                    loadmodel->surfaces[in->firstsurface].texinfo->texture->name[2] == 'y')
				{
                    out->children[j] = CONTENTS_SKY;
				}
                else
				{
					out->children[j] = child->contents;
				}
#else
                out->children[j] = child->contents;
#endif
            }
            else out->children[j] = child - loadmodel->nodes;
        }
    }
}


/*
=================
Mod_LoadMarksurfaces
=================
*/
void Mod_LoadMarksurfaces (lump_t *l)
{
    int		i, j, count;
    short		*in;
    glmsurface_t **out;

    in = (void *)(mod_base + l->fileofs);
    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->marksurfaces = out;
    loadmodel->nummarksurfaces = count;

    for ( i=0 ; i<count ; i++)
    {
        j = LittleShort(in[i]);
        if (j >= loadmodel->numsurfaces)
            Sys_Error ("Mod_ParseMarksurfaces: bad surface number");
        out[i] = loadmodel->surfaces + j;
    }
}


/*
=================
Mod_LoadSurfedges
=================
*/
void Mod_LoadSurfedges (lump_t *l)
{
    int		i, count;
    int		*in, *out;

    in = (void *)(mod_base + l->fileofs);
    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*sizeof(*out), loadname);

    loadmodel->surfedges = out;
    loadmodel->numsurfedges = count;

    for ( i=0 ; i<count ; i++)
        out[i] = LittleLong (in[i]);
}


/*
=================
Mod_LoadPlanes
=================
*/
void Mod_LoadPlanes (lump_t *l)
{
    int			i, j;
    mplane_t	*out;
    dplane_t 	*in;
    int			count;
    int			bits;

    in = (void *)(mod_base + l->fileofs);
    if (l->filelen % sizeof(*in))
        Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
    count = l->filelen / sizeof(*in);
    out = Hunk_AllocName ( count*2*sizeof(*out), loadname);

    loadmodel->planes = out;
    loadmodel->numplanes = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        bits = 0;
        for (j=0 ; j<3 ; j++)
        {
            out->normal[j] = LittleFloat (in->normal[j]);
            if (out->normal[j] < 0)
                bits |= 1<<j;
        }

        out->dist = LittleFloat (in->dist);
        out->type = LittleLong (in->type);
        out->signbits = bits;
    }
}

/*
=================
RadiusFromBounds
=================
*/
float RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
    int		i;
    vec3_t	corner;

    for (i=0 ; i<3 ; i++)
	{
        corner[i] = fabs(mins[i]) > fabs(maxs[i]) ? fabs(mins[i]) : fabs(maxs[i]);
	}
    return VLength (corner);
}

/*
=================
Mod_ParseEntityLump

some of this stuff is only done on the worldmodel, other stuff is done on all bsp models
(eg. the WAD)
=================
*/
void Mod_ParseEntityLump (lump_t *l)
{
	byte		*in;

	int			i, j;
	int			entnum;

	char		in_string[1024];
	char		last_name[1024];

	qboolean	in_entity;
	qboolean	reading_name;
	qboolean	reading_value;

	// stuff we need to store out for an entity - although the max value is 1024, most of
	// these are never going to be that big, so we can reduce the sizes accordingly.
	char		e_classname[100];
	int			e_worldtype;
	char		e_skybox[100];
	char		e_wad[1024];

	// initialize wadstring to a zero length string - same for skybox
	wadstring[0] = '\0';
	e_skybox[0] = '\0';

	if (!l->filelen)
	{
		// the bsp will not load anyway if there is no entities in it - we need at least
		// worldspawn and a start point - so we'll just do nothing and let quake handle it
		// in it's normal fashion (probably by crashing) - although the build tools should
		// catch all these problems anyway.
		loadmodel->entities = NULL;
		return;
	}
	in = mod_base + l->fileofs;

	entnum = 0;
	in_entity = false;

	reading_name = true;
	reading_value = false;

	// i is always the current position of the pointer into the entity lump - hence the fact
	// that i do some strange looking stuff with it below...
	for (i = 0; i < l->filelen;)
	{
		if (in_entity && *in == '"')
		{
			// read and store out a key/value pair - incrementing 'i' here is intentional!
			// no need for an end condition cos the build tools will have already tested the
			// maximum lengths for us
			for (j = 0;;)
			{
				in++;
				i++;
				if (*in == '"') break;
				in_string[j++] = *in;
			}
			in_string[j] = '\0';

			// only quake bsps have a worldtype key - this is not 100% reliable (especially
			// if using worldcraft) but is quite close - we can perform other tests later.
			if (!strcmp (in_string, "worldtype") && hlbsp)
			{
				hlbsp = false;
				qsg30 = true;
			}

			if (!worldtype_found)
			{
				// switch whether we're currently reading the name or the value
				if (reading_name)
				{
					reading_name = false;
					reading_value = true;

					// store out the name cos what we do with the value depends on it
					strcpy (last_name, in_string);
				}
				else if (reading_value)
				{
					reading_name = true;
					reading_value = false;

					// act according to the value of last_name - there is a limited amount of
					// stuff here we're actually interested in - but it can be expanded later
					if (!strcmp (last_name, "classname"))
					{
						strcpy (e_classname, in_string);
					}
					else if (!strcmp (last_name, "worldtype"))
					{
						e_worldtype = atoi (in_string);
					}
					else if (!strcmp (last_name, "_skybox"))
					{
						strcpy (e_skybox, in_string);
					}
					else if (!strcmp (last_name, "_skyname"))
					{
						strcpy (e_skybox, in_string);
					}
					else if (!strcmp (last_name, "skyname"))
					{
						strcpy (e_skybox, in_string);
					}
					else if (!strcmp (last_name, "wad"))
					{
						strcpy (e_wad, in_string);
					}
				}
			}
			else
			{
				// switch whether we're currently reading the name or the value
				if (reading_name)
				{
					reading_name = false;
					reading_value = true;

					// store out the name cos what we do with the value depends on it
					strcpy (last_name, in_string);
				}
				else if (reading_value)
				{
					reading_name = true;
					reading_value = false;

					// act according to the value of last_name - there is a limited amount of
					// stuff here we're actually interested in - but it can be expanded later
					// we also do wad here in case of hl bsp models
					if (!strcmp (last_name, "classname"))
					{
						strcpy (e_classname, in_string);
					}
					else if (!strcmp (last_name, "wad"))
					{
						strcpy (e_wad, in_string);
					}
				}
			}
		}

		if (*in == '{')
		{
			// start a new entity
			in_entity = true;

			e_wad[0] = '\0';

			if (!worldtype_found)
			{
				// initialize all the stuff we're saving out as zero length strings
				e_classname[0] = '\0';
				strcpy (e_skybox, "none");
				e_worldtype = 0;
			}
		}
		else if (*in == '}')
		{
			// finish current entity
			entnum++;
			in_entity = false;

			if (e_wad[0] != '\0')
			{
				// save out the wad string for hl wad3 loading
				strcpy (wadstring, e_wad);
			}

			if (!worldtype_found)
			{
				// set variables, etc depending on the entity - in practice, we'll only need
				// worldspawn - but this still does the entire entity lump in case we need to
				// expand it later...
				if (!strcmp (e_classname, "worldspawn"))
				{
					// worldtype
					global_worldtype = e_worldtype;

					// skybox
					Con_DPrintf ("Name is %s\n", e_skybox);

					// persistent skybox
					if (!strcmp (e_skybox, "none"))
					{
						if (_r_persist.value)
						{
							gl_useskybox = 1;
						}
						else
						{
							gl_useskybox = 0;
							Cvar_Set ("_r_skybox", "none");
						}
					}
					else if (e_skybox[0] == '\0')
					{
						gl_useskybox = 0;
						Cvar_Set ("_r_skybox", "none");
					}
					else
					{
						gl_useskybox = 1;
						Cvar_Set ("_r_skybox", e_skybox);
					}
				}
			}
		}
		in++;
		i++;
	}
	Con_DPrintf ("ParseEntityLump : %i Entities read from BSP\n", entnum);

	// we might as well load the entities while we're here...
	loadmodel->entities = Hunk_AllocName ( l->filelen, loadname);
	memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);
}

/*
=================
Mod_VerifyBSP30Type
=================
*/
void Mod_VerifyBSP30Type (lump_t *l)
{
    dmiptexlump_t	*m;
    miptex_t		*mt;
    int				i, j, pixels;

    // if there's no textures in the bsp, we'll let quake's builtin error checking take over
    // (ie. crash the engine)
    if (!l->filelen) return;

    m = (dmiptexlump_t *) (mod_base + l->fileofs);
    m->nummiptex = LittleLong (m->nummiptex);

    for (i = 0; i < m->nummiptex; i++)
    {
        m->dataofs[i] = LittleLong (m->dataofs[i]);

        if (m->dataofs[i] == -1) continue;

        mt = (miptex_t *) ((byte *) m + m->dataofs[i]);
        mt->width = LittleLong (mt->width);
        mt->height = LittleLong (mt->height);

        for (j = 0; j < MIPLEVELS; j++)
		{
            mt->offsets[j] = LittleLong (mt->offsets[j]);
		}

        // if the textures are in an external wad it's definitely halflife
        if (!mt->offsets[0]) return;

		// only halflife has '-' textures
        if (mt->name[0] == '-') return;

        // only quake has '*' textures
        if (mt->name[0] == '*') goto set_qsg30;

        pixels = mt->width * mt->height / 64 * 85;

        // if the pixels + 770 is greater than the size of the texlump, it's quake - this gives
        // 100% automatic identification reliability (if i have the maths correct.)
        if ((l->filelen + l->fileofs) < (l->fileofs + 770 + sizeof (miptex_t) + m->dataofs[i] + pixels))
		{
            goto set_qsg30;
		}
    }
    return;

// assume halflife format by default
set_qsg30:
    hlbsp = false;
    qsg30 = true;
    return;
}

/*
=================
Mod_LoadBrushModel
=================
*/
void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
    int			i, j;
    dheader_t	*header;
    dmodel_t 	*bm;

    loadmodel->type = mod_brush;

    header = (dheader_t *)buffer;

    i = LittleLong (header->version);

    if (i == 29)
    {
        hlbsp = false;
        qsg30 = false;
    }
    else if (i == 30)
    {
        // assume version 30 BSPs are halflife format until we prove otherwise.
        qsg30 = false;
        hlbsp = true;
    }
    else
    {
        // about time i implemented mr. g's tut
        //Sys_Error ("Invalid BSP Version %i\n", i);
        Con_Printf ("Invalid BSP Version %i\n", i);
        mod->numvertexes = -1;
        return;
    }
    mod_base = (byte *) header;

    // swap all the lumps
    for (i = 0; i < sizeof (dheader_t) / 4; i++)
	{
        ((int *) header)[i] = LittleLong (((int *) header)[i]);
	}

    // this gets stuff we need to know like skyboxes and the worldtype from the entities lump.
    // it also loads it normally.  we can get more info about any entity we want using this.
    // R_ParseEntityLump also determines if a version 30 BSP is halflife or qsg format.
    Mod_ParseEntityLump (&header->lumps[LUMP_ENTITIES]);

    if (hlbsp)
	{
		Mod_VerifyBSP30Type (&header->lumps[LUMP_TEXTURES]);
	}

#ifndef HALFLIFEBSP
    // if we don't support halflife BSPs we want to exit gracefully (crash for now...)
    if (hlbsp) Sys_Error ("Halflife BSP 30 format is not supported in this Quake engine");
#endif

    // the code in here is only executed when loading the worldmodel.
    if (!worldtype_found)
    {
        // halflife maps use a worldtype of base
        if (!hlbsp)
        {
            if (global_worldtype < 0 || global_worldtype > 2) global_worldtype = 0;
            Con_DPrintf ("Worldtype is %i\n", global_worldtype);

            // zer uses a global_worldtype of 3 for a red simple sky colour
            if (zer) global_worldtype = 3;
        }
        else global_worldtype = 2;

        worldtype_found = 1;

        // gl_useclight used to be set all over the bloody place - it's far handier to set
        // it in one place, and here is the most logical...
        if (gl_colouredlight.value)
		{
            gl_useclight = 1;
		}
        else
		{
			gl_useclight = 0;
		}

        // there's probably more I can add here... i've removed tjunctions because having them
        // in denies the player the option to set them manually
        if (gl_fastbutugly.value)
        {
            glHint (GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
            glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
            glHint (GL_FOG_HINT, GL_FASTEST);
            glHint (GL_POINT_SMOOTH_HINT, GL_FASTEST);
            glHint (GL_LINE_SMOOTH_HINT, GL_FASTEST);
        }
        else
        {
            glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);
            glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
            glHint (GL_FOG_HINT, GL_NICEST);
            glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
            glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
        }
    }

    // load into heap
    Mod_LoadVertexes (&header->lumps[LUMP_VERTEXES]);
    Mod_LoadEdges (&header->lumps[LUMP_EDGES]);
    Mod_LoadSurfedges (&header->lumps[LUMP_SURFEDGES]);
    Mod_LoadTextures (&header->lumps[LUMP_TEXTURES]);
    Mod_LoadLighting (&header->lumps[LUMP_LIGHTING]);
    Mod_LoadPlanes (&header->lumps[LUMP_PLANES]);
    Mod_LoadTexinfo (&header->lumps[LUMP_TEXINFO]);
    Mod_LoadFaces (&header->lumps[LUMP_FACES]);
    Mod_LoadMarksurfaces (&header->lumps[LUMP_MARKSURFACES]);
    Mod_LoadVisibility (&header->lumps[LUMP_VISIBILITY]);
    Mod_LoadLeafs (&header->lumps[LUMP_LEAFS]);
    Mod_LoadNodes (&header->lumps[LUMP_NODES]);
    Mod_LoadClipnodes (&header->lumps[LUMP_CLIPNODES]);
    Mod_LoadSubmodels (&header->lumps[LUMP_MODELS]);
    Mod_MakeHull0 ();

    mod->numframes = 2;		// regular and alternate animation

    // set up the submodels (FIXME: this is confusing)
    for (i=0 ; i<mod->numsubmodels ; i++)
    {
        bm = &mod->submodels[i];

        mod->hulls[0].firstclipnode = bm->headnode[0];

        for (j=1 ; j<MAX_MAP_HULLS ; j++)
        {
            mod->hulls[j].firstclipnode = bm->headnode[j];
            mod->hulls[j].lastclipnode = mod->numclipnodes-1;
        }

        mod->firstmodelsurface = bm->firstface;
        mod->nummodelsurfaces = bm->numfaces;

        VectorCopy (bm->maxs, mod->maxs);
        VectorCopy (bm->mins, mod->mins);

        mod->radius = RadiusFromBounds (mod->mins, mod->maxs);

        mod->numleafs = bm->visleafs;

        if (i < mod->numsubmodels-1)
        {
            // duplicate the basic information
            char	name[10];

            sprintf (name, "*%i", i+1);
            loadmodel = Mod_FindName (name);
            *loadmodel = *mod;
            strcpy (loadmodel->name, name);
            mod = loadmodel;
        }
    }
}

