/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// alias model scripting

#include "quakedef.h"

#ifdef USE_SCRIPTS
#include "script_common.h"

model_t *Mod_ForName (char *name, qboolean crash);

typedef enum
{
    fx_none,
    fx_missile,
    fx_torch,
    fx_powerup,
    fx_bolt
} fx_type_t;


// check a given script file for an entry relating to a given alias model and set the
// various effects, etc if there is one present.
void Alias_Script_Execute (char *scriptname, FILE *scriptfile)
{
    int i;
    char skey[64], value[64];
    char model_name[64];
    float trans;
    int is_glow;
    float glow_rgb[3];
    int shadow;
    int shading;
    fx_type_t fx_type;
    model_t *aliasmod;
    aliashdr_t *aliashdr;

    // read it
    while (1)
    {
        // set the defaults for the items
        trans = 1.0;
        is_glow = 0;
        glow_rgb[0] = 1.0;
        glow_rgb[1] = 1.0;
        glow_rgb[2] = 1.0;
        shadow = 0;
        shading = 0;
        fx_type = fx_none;

        // read until we get an opening brace - in this instance an eof could easily
        // signify that we've just hit the genuine eof, so we can finish up OK.
        if (!Script_ParseOpeningBrace (scriptfile)) goto finish_script;

        while (1)
        {
            // read the entries - the current character HAS to be a '"' so we can read away
            // happily, safe in the knowledge that whoever wrote the script did it correctly.
            if (!Script_Read (scriptfile, &skey[0])) goto bad_script;
            if (!Script_Read (scriptfile, &value[0])) goto bad_script;

            // depending on the value of skey, we do different things - i know i should be
            // using strcpy instead of sprintf, but i can never remember which arg goes where
            // in strcpy, and i'm too damn lazy to look it up...
            if (!strcmp (skey, "model"))
                sprintf (model_name, "%s", value);
            else if (!strcmp (skey, "trans"))
                trans = atof (value);
            else if (!strcmp (skey, "glow"))
            {
                is_glow = atoi (value);

                if (is_glow)
                {
                    // read 3 more values for the glow rgb
                    if (!Script_Read (scriptfile, value)) goto bad_script;
                    glow_rgb[0] = atof (value);

                    if (!Script_Read (scriptfile, value)) goto bad_script;
                    glow_rgb[1] = atof (value);

                    if (!Script_Read (scriptfile, value)) goto bad_script;
                    glow_rgb[2] = atof (value);
                }
            }
            else if (!strcmp (skey, "shadow"))
                shadow = atoi (value);
            else if (!strcmp (skey, "shading"))
                shading = atoi (value);
            else if (!strcmp (skey, "special"))
            {
                if (!strcmp (value, "missile"))
                    fx_type = fx_missile;
                else if (!strcmp (value, "torch"))
                    fx_type = fx_torch;
                else if (!strcmp (value, "powerup"))
                    fx_type = fx_powerup;
                else if (!strcmp (value, "bolt"))
                    fx_type = fx_bolt;
                else fx_type = fx_none;
            }

            // check the closing brace - if we have one and the current model name is the
            // one we want, then we store out the new stuff and continue
            i = Script_ParseClosingBrace (scriptfile);

            if (i == -1)
                goto bad_script;
            else if (i == 1)
            {
                // save out the fx
                aliasmod = Mod_ForName (model_name, false);

                if (!aliasmod)
                {
                    // add a path and an extension and try again
                    sprintf (model_name, "progs/%s.mdl", model_name);
                    aliasmod = Mod_ForName (model_name, false);

                    // no such model - so just skip to the next
                    if (!aliasmod) goto next_script_item;
                }

                // strictly speaking this should give an error, but we want to be a little
                // more lenient.  might make it harder to debug scripts, though...
                if (aliasmod->type != mod_alias) goto next_script_item;

                // save out the stuff
                aliashdr = (aliashdr_t *) Mod_Extradata (aliasmod);
                aliashdr->cast_shadow = shadow;
                aliashdr->is_glowmap = is_glow;
                aliashdr->glow_rgb[0] = glow_rgb[0];
                aliashdr->glow_rgb[1] = glow_rgb[1];
                aliashdr->glow_rgb[2] = glow_rgb[2];
                aliashdr->is_fullbright = shading;

                aliashdr->trans_level = trans;
                aliashdr->is_bolt = false;
                aliashdr->is_missile = false;
                aliashdr->is_powerup = false;
                aliashdr->is_torch = false;

                if (fx_type == fx_missile)
                    aliashdr->is_missile = true;
                else if (fx_type == fx_torch)
                    aliashdr->is_torch = true;
                else if (fx_type == fx_bolt)
                    aliashdr->is_bolt = true;
                else if (fx_type == fx_powerup)
                    aliashdr->is_powerup = true;
                else
                {
                    aliashdr->is_missile = false;
                    aliashdr->is_torch = false;
                    aliashdr->is_bolt = false;
                    aliashdr->is_powerup = false;
                }

                // done reading this entry so go to the next
                goto next_script_item;
            }
        }
next_script_item:
        ;
    }

bad_script:
    Con_Printf ("Error parsing scripts/%s.mhs\n", scriptname);

finish_script:
    return;
}


// mod command for changing individual alias model properties - ta to shadow for the
// inspiration
void Mod_f (void)
{
    char model_name[64];
    model_t *aliasmod;
    aliashdr_t *aliashdr;

    // mod <model_name> <property> <value(s)>
    if (Cmd_Argc () < 4)
    {
        Con_Printf ("mod <model_name> <property> <value(s)>: modify an alias model property\n");
        return;
    }

    // find the model - assume the full path and extension are given, if it can't be found then
    // add the default path and try again.
    aliasmod = Mod_ForName (Cmd_Argv (1), false);

    if (!aliasmod)
    {
        sprintf (model_name, "progs/%s.mdl", Cmd_Argv (1));
        aliasmod = Mod_ForName (model_name, false);

        if (!aliasmod)
        {
            // oh shit!
            Con_Printf ("Cannot find %s\n", Cmd_Argv (1));
            return;
        }
    }

    if (aliasmod->type != mod_alias)
    {
        // oh shit!
        Con_Printf ("%s is not an Alias Model\n", Cmd_Argv (1));
        return;
    }

    // save out the stuff
    aliashdr = (aliashdr_t *) Mod_Extradata (aliasmod);

    // what we do next depends on the property we try to set.
    if (!strcmp (Cmd_Argv (2), "trans"))
        aliashdr->trans_level = atof (Cmd_Argv (3));
    else if (!strcmp (Cmd_Argv (2), "glow"))
    {
        // yuck! - either 1 or 4 values to be set here!!!
        if (atoi (Cmd_Argv (3)) && Cmd_Argc () < 7)
        {
            // oh shit!
            Con_Printf ("You must specify 3 colour values\n");
            return;
        }

        if (!atoi (Cmd_Argv (3)) && Cmd_Argc () > 4)
        {
            // oh shit!
            Con_Printf ("You must not specify any colour values\n");
            return;
        }

        aliashdr->is_glowmap = atoi (Cmd_Argv (3));

        if (atoi (Cmd_Argv (3)))
        {
            aliashdr->glow_rgb[0] = atof (Cmd_Argv (4));
            aliashdr->glow_rgb[1] = atof (Cmd_Argv (5));
            aliashdr->glow_rgb[2] = atof (Cmd_Argv (6));
        }
    }
    else if (!strcmp (Cmd_Argv (2), "shadow"))
        aliashdr->cast_shadow = atoi (Cmd_Argv (3));
    else if (!strcmp (Cmd_Argv (2), "shading"))
        aliashdr->is_fullbright = atoi (Cmd_Argv (3));
    else if (!strcmp (Cmd_Argv (2), "special"))
    {
        // yuck!  again!  one of 4 different values
        aliashdr->is_missile = false;
        aliashdr->is_torch = false;
        aliashdr->is_bolt = false;
        aliashdr->is_powerup = false;

        if (!strcmp (Cmd_Argv (3), "missile"))
            aliashdr->is_missile = true;
        else if (!strcmp (Cmd_Argv (3), "torch"))
            aliashdr->is_torch = true;
        else if (!strcmp (Cmd_Argv (3), "powerup"))
            aliashdr->is_powerup = true;
        else if (!strcmp (Cmd_Argv (3), "bolt"))
            aliashdr->is_bolt = true;
    }
}
#endif

